/*
 * Compiz Tiles Plugin
 *
 * tiles.cpp
 *
 * Copyright (c) 2011 Sam Spilsbury <smspillaz@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include "animatedrects.h"

namespace
{
    const short OPAQUE = std::numeric_limits<short>::max ();
}

class PrivateLinearAnimatedRect
{
    public:

	boost::function<CompRect ()> mGetFromRect;
	boost::function<CompRect ()> mGetToRect;

	AnimatableRectInterface::Type mType;
	float mProgress;
};

CompRect
scaleChild (const CompRect &child, const CompRect &parent, const CompRect &newParent)
{
    /* Next, calculate translation factor due to change in scale
     * from origin */
    CompPoint childOriginOffset (child.pos () - parent.pos ());

    /* Calculate scale factor from parent to newParent */
    float xScale = newParent.width () / static_cast <float> (parent.width ());
    float yScale = newParent.height () / static_cast <float> (parent.height ());

    /* Now scale the childOriginOffset by the scale factor, this is
     * the new offset from the parent origin */
    float xTranslate = childOriginOffset.x () * xScale;
    float yTranslate = childOriginOffset.y () * yScale;

    /* Now apply these transformations */
    return CompRect (newParent.x () + xTranslate,
		     newParent.y () + yTranslate,
		     child.width () * xScale,
		     child.height () * yScale);
}

LinearAnimatedRect::LinearAnimatedRect () :
    priv (new PrivateLinearAnimatedRect)
{
    priv->mType = AnimatableRectInterface::Type::Complete;
    priv->mProgress = 1.0f;
}

void
LinearAnimatedRect::setFromRectFunc (const boost::function<CompRect ()> &f)
{
    priv->mGetFromRect = f;
}

void
LinearAnimatedRect::setToRectFunc (const boost::function<CompRect ()> &f)
{
    priv->mGetToRect = f;
}

bool
LinearAnimatedRect::setDirection (AnimatableRectInterface::Type type)
{
    if (priv->mType != type)
    {
	/* Resume from last position when changing from
	 * complete to either in or out, or from in to out
	 * and vice versa */

	if ((priv->mType == AnimatableRectInterface::Type::Complete ||
	     priv->mType == AnimatableRectInterface::Type::Out ||
	     priv->mType == AnimatableRectInterface::Type::In) &&
	    (type == AnimatableRectInterface::Type::In ||
	     type == AnimatableRectInterface::Type::Out))
	    priv->mProgress = 1.0f - priv->mProgress;
	else if (type == AnimatableRectInterface::Type::Complete)
	    priv->mProgress = 1.0f;

	priv->mType = type;
	return true;
    }

    return false;
}

const unsigned int anim_time = 300;

bool
LinearAnimatedRect::progressAnimation (int ms, bool force)
{
    if (priv->mType != AnimatableRectInterface::Type::Complete ||
	force)
    {
        //float chunk = ms / static_cast <float> (anim_time);
	priv->mProgress += ms / static_cast <float> (anim_time);

	if (fabsf (ms))
	{
	    if (priv->mProgress >= 1.0f)
	    {
		priv->mProgress = 1.0f;
		priv->mType = AnimatableRectInterface::Type::Complete;
	    }
	    else if (priv->mProgress <= 0.0f)
	    {
		priv->mProgress = 0.0f;
		priv->mType = AnimatableRectInterface::Type::Complete;

	    }
	}
    }

    return (priv->mType != AnimatableRectInterface::Type::Complete) || force;
}

CompRect
LinearAnimatedRect::tweenedGeometry (const CompRect &rect)
{
    CompRect from (priv->mGetFromRect.empty () ? rect : priv->mGetFromRect ());
    CompRect to   (priv->mGetToRect.empty () ? rect : priv->mGetToRect ());

    CompRect current (from.x () + ((to.x () - from.x ()) * priv->mProgress),
		      from.y () + ((to.y () - from.y ()) * priv->mProgress),
		      from.width () + ((to.width () - from.width ()) * priv->mProgress),
		      from.height () + ((to.height () - from.height ()) * priv->mProgress));

    return current;
}

unsigned int
LinearAnimatedRect::getOpacity ()
{
    return OPAQUE * priv->mProgress;
}
