/*
 * Compiz Tiles Plugin
 *
 * tiles.cpp
 *
 * Copyright (c) 2011 Sam Spilsbury <smspillaz@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include "fbeffect.h"
#include <boost/assign/std/vector.hpp>
#include <boost/math/distributions/normal.hpp>
#include <boost/make_shared.hpp>

#include <composite/composite.h>

using namespace boost::assign;
using boost::math::normal;

namespace GL
{

    GLUseProgramProc glUseProgramP;
    GLGetUniformLocationProc glGetUniformLocationP;
    GLUniform1fProc glUniform1fP;
    GLUniform2fProc glUniform2fP;
    GLUniform1fvProc glUniform1fvP;
    GLUniform1iProc glUniform1iP;
    GLValidateProgramProc glValidateProgramP;
    GLGetProgramivProc glGetProgramivP;
    GLGetProgramInfoLogProc glGetProgramInfoLogP;

    GLCreateProgramProc glCreateProgramP;
    GLCreateShaderProc  glCreateShaderP;
    GLShaderSourceProc  glShaderSourceP;
    GLCompileShaderProc glCompileShaderP;
    GLAttachShaderProc  glAttachShaderP;
    GLLinkProgramProc   glLinkProgramP;
    GLGetShaderInfoLogProc glGetShaderInfoLogP;


}

struct FramebufferDrawVertexData
{
    CompRegion geometry;
    CompRegion clipRegion;
};

class PrivateFramebufferPainter
{
    public:

	PrivateFramebufferPainter (const CompRegion &clip,
				   const CompRect &windowRelativeRect) :
	    mClipRegion (clip),
	    mWindowRelativeRect (windowRelativeRect)
	{
	}

	BackingFramebufferInterface::Ptr mBacking;
    	GLint     mPreviousProgram;
    	GLfloat   mProgress;
	GLMatrix  *mTransform;
	CompOutput *mOutput;
	GLfloat    mAlpha;
	CompRect mGeometry;
	const CompRegion &mClipRegion;
	const CompRect   &mWindowRelativeRect;
};

class PrivateBackingFramebuffer
{
    public:

	std::auto_ptr <GLFramebufferObject> mCompizFBO;
	CompSize                            mSize;

	/* Don't delete, not owned by us */
	GLProgram                           *mProgram;

	std::vector <GLfloat> mWeights;

	GLWindow *mGLWindow;
};

class RebindableGLFramebufferObject :
    public RebindableBackingFramebufferInterface
{
    public:

	typedef boost::shared_ptr <RebindableGLFramebufferObject> Ptr;

	RebindableGLFramebufferObject (GLFramebufferObject *obj,
				       const CompRect      &oldViewport);

	void rebind ();

    private:

	std::auto_ptr <PrivateRebindableGLFramebufferObject> priv;
};

class PrivateRebindableGLFramebufferObject
{
    public:

	GLFramebufferObject *mFbo;
	CompRect            mOldViewport;
};

RebindableGLFramebufferObject::RebindableGLFramebufferObject (GLFramebufferObject *fbo,
							      const CompRect      &oldViewport) :
    priv (new PrivateRebindableGLFramebufferObject)
{
    priv->mFbo = fbo;
    priv->mOldViewport = oldViewport;
}

void
RebindableGLFramebufferObject::rebind ()
{
    GLFramebufferObject::rebind (priv->mFbo);
    glViewport (priv->mOldViewport.x (),
		priv->mOldViewport.y (),
		priv->mOldViewport.width (),
		priv->mOldViewport.height ());
}

BlurBackingFramebuffer::Ptr
BlurBackingFramebuffer::create (const CompSize &sz,
				GLProgram      *program,
				GLWindow       *window)
{
    return BlurBackingFramebuffer::Ptr (new BlurBackingFramebuffer (sz, program, window));
}

BlurBackingFramebuffer::BlurBackingFramebuffer (const CompSize &sz,
						GLProgram      *program,
						GLWindow       *window) :
    priv (new PrivateBackingFramebuffer ())
{
    priv->mCompizFBO.reset (new GLFramebufferObject ());
    priv->mSize = sz;
    priv->mProgram = program;
    priv->mGLWindow = window;

    if (!priv->mCompizFBO->allocate (priv->mSize))
	throw std::exception ();
}

bool
BlurBackingFramebuffer::drawAtPosition (FramebufferDrawVertexData *inVertexData,
					GLMatrix                  *matrix,
					const CompRect            &windowRelativeRect)
{
    GLTexture               *texture = priv->mCompizFBO->tex ();
    GLTexture::Matrix       texmatrix (texture->matrix ());
    GLTexture::MatrixList   matl;

    CompRect r (inVertexData->geometry.boundingRect ());

    texmatrix.xx *= windowRelativeRect.width () / static_cast <float> (r.width ());
    texmatrix.yy *= windowRelativeRect.height () / static_cast <float> (r.height ());

    texmatrix.x0 -= texmatrix.xx * r.x1 ();
    texmatrix.y0 -= texmatrix.yy * r.y1 ();

    matl.push_back (texmatrix);

    GLVertexBuffer *streamingBuffer = priv->mGLWindow->vertexBuffer ();
    streamingBuffer->begin ();

    priv->mGLWindow->glAddGeometry (matl,
				    inVertexData->geometry,
				    inVertexData->clipRegion);

    if (streamingBuffer->end ())
    {
	streamingBuffer->addUniform2f ("TextureSize0", texture->width (), texture->height ());
	streamingBuffer->addUniform ("W0", priv->mWeights[0]);
	streamingBuffer->addUniform ("W1", priv->mWeights[1]);
	streamingBuffer->addUniform ("W2", priv->mWeights[2]);
	streamingBuffer->addUniform ("W3", priv->mWeights[3]);
	streamingBuffer->addUniform ("W4", priv->mWeights[4]);
	streamingBuffer->addUniform ("W5", priv->mWeights[5]);
	streamingBuffer->addUniform ("W6", priv->mWeights[6]);

	GLWindowPaintAttrib wAttrib;

	wAttrib.opacity = OPAQUE;
	wAttrib.brightness = BRIGHT;
	wAttrib.saturation = COLOR;

	streamingBuffer->setProgram (priv->mProgram);
	priv->mGLWindow->glDrawTexture (texture,
					*matrix,
					wAttrib,
					PAINT_WINDOW_BLEND_MASK |
					PAINT_WINDOW_TRANSLUCENT_MASK |
					PAINT_WINDOW_TRANSFORMED_MASK);
	streamingBuffer->setProgram (NULL);
	return true;
    }
    else
	return false;
}

RebindableBackingFramebufferInterface::Ptr
BlurBackingFramebuffer::bind (const CompRect &viewport)
{
    if (!priv->mCompizFBO->checkStatus ())
	throw std::exception ();

    GLint oldViewport[4];

    glGetIntegerv (GL_VIEWPORT, &oldViewport[0]);

    RebindableBackingFramebufferInterface::Ptr lastBacking (
		boost::make_shared <RebindableGLFramebufferObject> (priv->mCompizFBO->bind (),
								    CompRect (oldViewport[0],
									      oldViewport[1],
									      oldViewport[2],
									      oldViewport[3])));

    glViewport (0,
		0,
		viewport.width (),
		viewport.height ());
    glClearColor (0, 0, 0, 0.0);
    glClear (GL_COLOR_BUFFER_BIT);

    return lastBacking;
}

void
BlurBackingFramebuffer::unbind (const RebindableBackingFramebufferInterface::Ptr &restore)
{
    restore->rebind ();
}

void
BlurBackingFramebuffer::setWeights (const std::vector<GLfloat> &weights)
{
    priv->mWeights = weights;
}

FramebufferPainter::FramebufferPainter (const BackingFramebufferInterface::Ptr &backing,
					const CompOutput &output,
					GLfloat progress,
					GLMatrix &transform,
					const CompRect &geometry,
					const CompRect &windowRelativeRect,
					GLfloat  alpha,
					const CompRegion &clipRegion) :
    priv (new PrivateFramebufferPainter (clipRegion, windowRelativeRect))
{
    if (!backing)
    	throw std::exception ();

    priv->mBacking = backing;
    priv->mOutput = const_cast <CompOutput *> (&output);
    priv->mProgress = progress;
    priv->mTransform = &transform;
    priv->mAlpha = alpha;
    priv->mGeometry = geometry;
}

FramebufferPainter::~FramebufferPainter ()
{
    std::vector <GLfloat> Wdata;

    normal distrib (7 * (priv->mProgress), 1.5f + 40.0f * (1 - pow (priv->mProgress, 2)));
    double n = 7;
    double step = 1;
    double total = 0.0;
    for (double z = 0; z < step; z += (step / n))
    {
	total += boost::math::pdf (distrib, z);
	Wdata += boost::math::pdf (distrib, z);
    }

    /* Normalize */
    for (GLfloat &f : Wdata)
    {
	f *= (1.0f / total);
	f *= priv->mAlpha;
    }

    FramebufferDrawVertexData data;
    data.clipRegion = priv->mClipRegion;
    data.geometry = CompRegion (priv->mGeometry);

    priv->mBacking->setWeights (Wdata);
    priv->mBacking->drawAtPosition (&data, priv->mTransform, priv->mWindowRelativeRect);
}

BlurBackingFramebuffer::~BlurBackingFramebuffer ()
{
}

const CompSize &
BlurBackingFramebuffer::size () const
{
    return priv->mSize;
}

