/*
 * Compiz Tiles Plugin
 *
 * tiles.cpp
 *
 * Copyright (c) 2011 Sam Spilsbury <smspillaz@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include "fbmanager.h"

class PrivateFramebufferBinder
{
    public:

	BackingFramebufferInterface::Ptr           mBacking;
	RebindableBackingFramebufferInterface::Ptr mPreviousBufferObject;
	CompOutput *mOutput;

};

class PrivateUpdatableFramebufferCommunicator
{
    public:

	FramebufferUpdateQueueInterface *mUpdateQueue;
	FramebufferUpdateInterface *mUpdatable;
};

class PrivateUpdatableFramebufferForwarder
{
    public:

	UpdatableFramebufferForwarder::UpdateFunc mUpdateFunc;
};

class PrivateFramebufferUpdateQueue
{
    public:

	std::vector <FramebufferUpdateInterface *> mPendingFramebufferUpdates;
};

UpdatableFramebufferCommunicator::UpdatableFramebufferCommunicator (FramebufferUpdateQueueInterface *uq,
								    FramebufferUpdateInterface *ui) :
    priv (new PrivateUpdatableFramebufferCommunicator ())
{
    priv->mUpdateQueue = uq;
    priv->mUpdatable = ui;
}

UpdatableFramebufferForwarder::UpdatableFramebufferForwarder (const UpdateFunc &f) :
    priv (new PrivateUpdatableFramebufferForwarder ())
{
    priv->mUpdateFunc = f;
}

bool
UpdatableFramebufferForwarder::fboUpdate ()
{
    return priv->mUpdateFunc ();
}

void
UpdatableFramebufferCommunicator::setFramebufferHolder (FramebufferUpdateQueueInterface *uq)
{
    priv->mUpdateQueue = uq;
}

void
UpdatableFramebufferCommunicator::queueUpdate ()
{
    priv->mUpdateQueue->addQueuedUpdate (priv->mUpdatable);
}

void
UpdatableFramebufferCommunicator::dequeueUpdates ()
{
    priv->mUpdateQueue->removeUpdatesFromQueue (priv->mUpdatable);
}

FramebufferBinder::FramebufferBinder (const BackingFramebufferInterface::Ptr &backing, const CompOutput &output) :
    priv (new PrivateFramebufferBinder ())
{
    if (!backing)
	throw std::exception ();

    priv->mBacking = backing;
    priv->mOutput = const_cast <CompOutput *> (&output);

    priv->mPreviousBufferObject = priv->mBacking->bind (output);
}

FramebufferBinder::~FramebufferBinder ()
{
    priv->mBacking->unbind (priv->mPreviousBufferObject);

    delete priv;
}

FramebufferUpdateQueue::FramebufferUpdateQueue () :
    priv (new PrivateFramebufferUpdateQueue ())
{
}

bool
FramebufferUpdateQueue::addQueuedUpdate (FramebufferUpdateInterface *update)
{
    if (std::find (priv->mPendingFramebufferUpdates.begin (),
		   priv->mPendingFramebufferUpdates.end (),
		   update) != priv->mPendingFramebufferUpdates.end ())
	return false;

    priv->mPendingFramebufferUpdates.push_back (update);

    return true;
}

void
FramebufferUpdateQueue::removeUpdatesFromQueue (FramebufferUpdateInterface *update)
{
    std::vector <FramebufferUpdateInterface *>::iterator it = std::find (priv->mPendingFramebufferUpdates.begin (),
									 priv->mPendingFramebufferUpdates.end (),
									 update);

    if (it == priv->mPendingFramebufferUpdates.end ())
	return;

    priv->mPendingFramebufferUpdates.erase (it);
}

void
FramebufferUpdateQueue::dispatch ()
{
    for (FramebufferUpdateInterface *update : priv->mPendingFramebufferUpdates)
    	update->fboUpdate ();

    priv->mPendingFramebufferUpdates.clear ();
}

bool
FramebufferUpdateQueue::pending ()
{
    return !priv->mPendingFramebufferUpdates.empty ();
}

