/*
 * Compiz Tiles Plugin
 *
 * tile.h
 *
 * Copyright (c) 2011 Sam Spilsbury <smspillaz@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#ifndef _TILES_TILE_H
#define _TILES_TILE_H

#include <boost/shared_ptr.hpp>

#include <core/core.h>
#include <opengl/opengl.h>

#include "dnd.h"
#include "resizeallocator.h"
#include "animatedrects.h"
#include "drawable.h"

class TileAssignableInterface;
class TileLayerMiscInterface;
class TileLayerLockInterface;
class PrivateTile;
class MaximizationStateApplicationInterface;

class TileInterface :
    public DroppableInterface,
    public DraggableInterface,
    public SnapRemovableInterface,
    public ResizeAllocationInterface,
    public ResizeTargetInterface,
    public DropCommitInterface
{
    public:

	typedef boost::shared_ptr <TileInterface> Ptr;

	virtual CompRect current (const CompRect &rect) = 0;
	virtual ResizableInterface * window () = 0;
	virtual TileAssignableInterface * assigned () = 0;
	virtual bool allocateWindow (ResizableInterface *, TileAssignableInterface *) = 0;
	virtual bool setAnimation (const AnimatableRectInterface::Type &) = 0;
	virtual bool lockTargetSize () = 0;
	virtual void resized (CompRect &) = 0;
	virtual bool animate (int ms, bool force = false) = 0;
	virtual void damage () = 0;
	virtual bool associate (TileLayerMiscInterface *p) = 0;
	virtual void disassociate () = 0;

	virtual TileLayerMiscInterface * layer () = 0;

	virtual unsigned int opacity () = 0;
	virtual GLVector scale (const CompRect &r) = 0;
	virtual GLVector translation (const CompRect &r) = 0;

	virtual CompRect getDrawWindowGeometry () = 0;
	virtual CompRect getGrabGeometry () = 0;
};

class Tile :
    public TileInterface

{
    public:

	Tile (TileLayerMiscInterface *parent,
	      const boost::shared_ptr <AnimatableRectInterface> &,
	      DrawableWindowInterface *dwi,
	      const boost::shared_ptr <DroppableInterface> &,
	      const boost::shared_ptr <DraggableInterface> &,
	      const boost::shared_ptr <SnapRemovableInterface> &,
	      const boost::shared_ptr <ResizeAllocationInterface> &,
	      TileLayerLockInterface *,
	      MaximizationStateApplicationInterface * maximizable,
	      const boost::function<bool ()> &dropAction);
	Tile (TileLayerMiscInterface *parent,
	      const boost::shared_ptr <AnimatableRectInterface> &,
	      DrawableWindowInterface *dwi,
	      const boost::shared_ptr <DroppableInterface> &,
	      const boost::shared_ptr <DraggableInterface> &,
	      const boost::shared_ptr <SnapRemovableInterface> &,
	      const boost::shared_ptr <ResizeAllocationInterface> &,
	      TileLayerLockInterface *,
	      MaximizationStateApplicationInterface * maximizable);
	~Tile ();

	/* Drag and Drop */
	void setPotentialDropZone (const CompRect &);
	const CompRect & dropGeometry ();

	void dropCommit ();

	/* Target size */
	CompRect     getDrawWindowGeometry ();
	CompRect     getGrabGeometry ();
	const CompRect    & target ();
	const CompRect    & allocation ();
	CompRect     current (const CompRect &rect);
	ResizableInterface * window ();
	TileAssignableInterface * assigned ();
	unsigned int opacity ();
	GLVector     scale (const CompRect &r);
	GLVector     translation (const CompRect &r);

	/* Returns false if window cannot be put
	 * in this tile's allocation */
	bool tileableFitsAllocation (ResizableInterface *tileable);
	bool allocateWindow (ResizableInterface *, TileAssignableInterface *);

	/* setTargetSize will allocate a new target size
	 * for this tile, which will be animated to, while
	 * lockTargetSize actually resizes the window to that
	 * size */
	void setTarget (const CompRect &);
	bool setAllocation (const CompRect &);
	bool setAnimation (const AnimatableRectInterface::Type &);
	bool lockTargetSize ();

	void resized (CompRect &);

	bool animate (int ms, bool force = false);
	void damage ();

	void grab ();
	void applyDrag(const CompPoint &p);
	bool shouldSnapOut ();
	void ungrab ();

	bool associate (TileLayerMiscInterface *p);
	void disassociate ();

	TileLayerMiscInterface * layer ();

	void setGetSnapGeometryFunc (const SnapRemovableInterface::GetGeometryFunc &func);

	typedef boost::shared_ptr<Tile> Ptr;

    private:

	void init (TileLayerMiscInterface *parent,
		   const boost::shared_ptr <AnimatableRectInterface> &,
		   DrawableWindowInterface *dwi,
		   const boost::shared_ptr <DroppableInterface> &,
		   const boost::shared_ptr <DraggableInterface> &,
		   const boost::shared_ptr <SnapRemovableInterface> &,
		   const boost::shared_ptr <ResizeAllocationInterface> &,
		   TileLayerLockInterface *,
		   MaximizationStateApplicationInterface * maximizable,
		   const boost::function<bool ()> &dropAction);

	PrivateTile *priv;
};

#endif
