/*
 * Compiz Tiles Plugin
 *
 * tiles.cpp
 *
 * Copyright (c) 2011 Sam Spilsbury <smspillaz@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include <boost/shared_ptr.hpp>

#include "tile.h"
#include "dnd.h"
#include "resizeallocator.h"
#include "tilelayerlock.h"
#include "maximization.h"

class PrivateTile
{
    public:

	TileLayerMiscInterface *mParent;
	CompRect   mTarget;
	CompRect   mAllocation;
	ResizableInterface *mWindow;
	MaximizationStateApplicationInterface *mMaximizable;
	DrawableWindowInterface *mPaintable;
	TileAssignableInterface *mAssignable;
	TileLayerLockInterface *mTileLayerLock;
	CompRect mSaveGeometry;
	unsigned int         mSaveMask;

	boost::shared_ptr <DraggableInterface> mDraggable;
	boost::shared_ptr <DroppableInterface> mDroppable;
	boost::shared_ptr <SnapRemovableInterface> mSnapRemovable;
	boost::shared_ptr <AnimatableRectInterface> mAnimatableRect;
	boost::shared_ptr <ResizeAllocationInterface> mResizeAllocator;

	boost::function <bool ()> mDropAction;
};

void
Tile::init (TileLayerMiscInterface *parent,
	    const boost::shared_ptr <AnimatableRectInterface> &animatableRect,
	    DrawableWindowInterface *dwi,
	    const boost::shared_ptr <DroppableInterface> &droppable,
	    const boost::shared_ptr <DraggableInterface> &draggable,
	    const boost::shared_ptr <SnapRemovableInterface> &snapRemovable,
	    const boost::shared_ptr <ResizeAllocationInterface> &resizeAllocator,
	    TileLayerLockInterface *tileLayerLock,
	    MaximizationStateApplicationInterface *maximizable,
	    const boost::function <bool ()> &dropAction)
{
    priv = new PrivateTile;
    priv->mParent = parent;
    priv->mWindow = NULL;
    priv->mAnimatableRect = animatableRect;
    priv->mPaintable = dwi;
    priv->mDraggable = draggable;
    priv->mDroppable = droppable;
    priv->mSnapRemovable = snapRemovable;
    priv->mResizeAllocator = resizeAllocator;
    priv->mTileLayerLock = tileLayerLock;
    priv->mMaximizable = maximizable;
    priv->mDropAction  = dropAction;
}

Tile::Tile (TileLayerMiscInterface *parent,
	    const boost::shared_ptr <AnimatableRectInterface> &animatableRect,
	    DrawableWindowInterface *dwi,
	    const boost::shared_ptr <DroppableInterface> &droppable,
	    const boost::shared_ptr <DraggableInterface> &draggable,
	    const boost::shared_ptr <SnapRemovableInterface> &snapRemovable,
	    const boost::shared_ptr <ResizeAllocationInterface> &resizeAllocator,
	    TileLayerLockInterface *tileLayerLock,
	    MaximizationStateApplicationInterface *maximizable,
	    const boost::function <bool ()> &dropAction)
{
    init (parent,
	  animatableRect,
	  dwi,
	  droppable,
	  draggable,
	  snapRemovable,
	  resizeAllocator,
	  tileLayerLock,
	  maximizable,
	  dropAction);
}

Tile::Tile (TileLayerMiscInterface *parent,
	    const boost::shared_ptr <AnimatableRectInterface> &animatableRect,
	    DrawableWindowInterface *dwi,
	    const boost::shared_ptr <DroppableInterface> &droppable,
	    const boost::shared_ptr <DraggableInterface> &draggable,
	    const boost::shared_ptr <SnapRemovableInterface> &snapRemovable,
	    const boost::shared_ptr <ResizeAllocationInterface> &resizeAllocator,
	    TileLayerLockInterface *tileLayerLock,
	    MaximizationStateApplicationInterface *maximizable)
{
    init (parent,
	  animatableRect,
	  dwi,
	  droppable,
	  draggable,
	  snapRemovable,
	  resizeAllocator,
	  tileLayerLock,
	  maximizable,
	  /* The default action when dropping will just be to
	   * lock in place the target size */
	  boost::bind (&TileInterface::lockTargetSize, this));
}


Tile::~Tile ()
{
    if (priv->mWindow)
    {
	float xScale, yScale;

	xScale = priv->mSaveGeometry.width () / static_cast <float> (priv->mWindow->windowGeometry ().width ());
	yScale = priv->mSaveGeometry.height () / static_cast <float> (priv->mWindow->windowGeometry ().height ());

	CompWindow::Geometry resizeTarget (pointerX - ((pointerX - priv->mWindow->windowGeometry ().x ()) * xScale),
					   pointerY - ((pointerY - priv->mWindow->windowGeometry ().y ()) * yScale),
					   priv->mSaveMask & CWWidth ? priv->mSaveGeometry.width () : priv->mWindow->windowGeometry ().width (),
					   priv->mSaveMask & CWHeight ? priv->mSaveGeometry.height () : priv->mWindow->windowGeometry ().height (), 0);

	priv->mWindow->applyResizeRules (false);
	priv->mWindow->resizeWindow (resizeTarget);
    }

    delete priv;
}

CompRect
Tile::getDrawWindowGeometry ()
{
    return priv->mPaintable->getDrawWindowGeometry ();
}

CompRect
Tile::getGrabGeometry ()
{
    return priv->mWindow->grabGeometry ();
}

const CompRect &
Tile::target ()
{
    return priv->mTarget;
}

const CompRect &
Tile::allocation ()
{
    return priv->mResizeAllocator->allocation ();
}

ResizableInterface *
Tile::window ()
{
    return priv->mWindow;
}

bool
Tile::tileableFitsAllocation (ResizableInterface *tileable)
{
    if (priv->mWindow)
	return false;

    return priv->mResizeAllocator->tileableFitsAllocation (tileable);
}

bool
Tile::allocateWindow (ResizableInterface *tw, TileAssignableInterface *ta)
{
    if (!tileableFitsAllocation (tw))
	return false;

    priv->mWindow = tw;
    priv->mAssignable = ta;
    priv->mSaveGeometry = priv->mWindow->windowGeometry ();
    priv->mSaveMask = CWWidth | CWHeight;

    return true;
}

bool
Tile::animate (int ms, bool force)
{
    if (ms || force)
	damage ();

    bool ret = priv->mAnimatableRect->progressAnimation (ms, force);

    if (ret)
	damage ();

    return ret;
}

bool
Tile::lockTargetSize ()
{
    priv->mAnimatableRect->setDirection (AnimatableRectInterface::Type::Complete);
    priv->mAnimatableRect->progressAnimation (0, true);
    CompRect windowGeometry (priv->mWindow->windowGeometry ());
    CompRect resizeTarget = priv->mAnimatableRect->tweenedGeometry (windowGeometry);

    resizeTarget = priv->mWindow->toWindowGeometry (resizeTarget);

    priv->mWindow->applyResizeRules (true);
    priv->mMaximizable->applyMaximizeState (resizeTarget);
    priv->mWindow->resizeWindow (resizeTarget);

    return true;
}

void
Tile::resized (CompRect &r)
{
    priv->mWindow->toWindowGeometry (r);
}

CompRect
Tile::current (const CompRect &rect)
{
    return priv->mAnimatableRect->tweenedGeometry (rect);
}

unsigned int
Tile::opacity ()
{
    return priv->mAnimatableRect->getOpacity ();
}

GLVector
Tile::scale (const CompRect &rect)
{
    GLVector v;
    CompRect cur = current (rect);

    v[GLVector::x] = cur.width () / static_cast <float> (rect.width ());
    v[GLVector::y] = cur.height () / static_cast <float> (rect.height ());
    v[GLVector::z] = 1.0f;
    v[GLVector::w] = 1.0f;

    return v;
}

GLVector
Tile::translation (const CompRect &rect)
{
    GLVector v;
    CompRect cur = current (rect);

    v[GLVector::x] = cur.x () - rect.x ();
    v[GLVector::y] = cur.y () - rect.y ();
    v[GLVector::z] = 0.0f;
    v[GLVector::w] = 0.0f;

    return v;
}

void
Tile::setTarget (const CompRect &r)
{
    priv->mTarget = r;
}

bool
Tile::setAllocation (const CompRect &r)
{
    return priv->mResizeAllocator->setAllocation (r);
}

bool
Tile::setAnimation (const AnimatableRectInterface::Type &type)
{
    /* Always bind fptrs */

    switch (type)
    {
	case AnimatableRectInterface::Type::In:
	    priv->mAnimatableRect->setFromRectFunc (boost::bind (&Tile::getDrawWindowGeometry, this));
	    priv->mAnimatableRect->setToRectFunc (boost::bind (&Tile::target, this));
	    break;
	case AnimatableRectInterface::Type::Out:
	    priv->mAnimatableRect->setFromRectFunc (boost::bind (&Tile::target, this));
	    priv->mAnimatableRect->setToRectFunc (boost::bind (&Tile::getDrawWindowGeometry, this));
	    break;
	case AnimatableRectInterface::Type::Complete:
	    priv->mAnimatableRect->setToRectFunc (boost::bind (&Tile::target, this));
	    priv->mAnimatableRect->setFromRectFunc (boost::bind (&Tile::getDrawWindowGeometry, this));
	default:
	    break;
    }

    damage ();
    bool ret = priv->mAnimatableRect->setDirection (type);
    return ret;

}

void
Tile::grab ()
{
    priv->mDraggable->grab ();
}

void
Tile::setPotentialDropZone (const CompRect &dz)
{
    priv->mDroppable->setPotentialDropZone (dz);
}

const CompRect &
Tile::dropGeometry ()
{
    return priv->mDroppable->dropGeometry ();
}

bool
Tile::shouldSnapOut ()
{
    return priv->mSnapRemovable->shouldSnapOut ();
}

void
Tile::dropCommit ()
{
    priv->mDropAction ();
    ungrab ();
}

void
Tile::applyDrag (const CompPoint &p)
{
    bool animationComplete = priv->mAnimatableRect->progressAnimation (0, false);
    bool doAnimate = !animationComplete && priv->mParent;

    const CompRect &dropZone = priv->mDroppable->dropGeometry ();
    const CompRect &allocation = priv->mResizeAllocator->allocation ();

    priv->mDraggable->applyDrag (p);

    /* Left or right ... */
    if (dropZone.height () > dropZone.width ())
    {
	/* Tiling to the right */
	if (dropZone.left () > allocation.left ())
	{
	    if (doAnimate)
		animate (p.x () * 3.0, true);
	}
	/* Tiling to the left */
	else
	{
	    if (doAnimate)
		animate (-p.x () * 3.0, true);
	}
    }
    else
    {
	/* Tiling to the bottom */
	if (dropZone.top () > allocation.top ())
	{
	    if (doAnimate)
		animate (p.y () * 3.0, true);
	}
	/* Tiling to the top */
	else
	{
	    if (doAnimate)
		animate (-p.y () * 3.0, true);
	}
    }
}

void
Tile::ungrab ()
{
    priv->mDraggable->ungrab ();
}

bool
Tile::associate (TileLayerMiscInterface *p)
{
    if (!priv->mParent)
    {
	priv->mParent = p;
	return true;
    }
    else
    {
	compLogMessage ("tiles", CompLogLevelWarn, "attempted to associate a tile with a layer \n" \
			" that was already associated!");
	return false;
    }
}

void
Tile::disassociate ()
{
    CompRect r = priv->mSaveGeometry;
    priv->mParent = NULL;


    float xScale = priv->mSaveGeometry.width () / static_cast <float> (priv->mWindow->windowGeometry ().width ());
    float yScale = priv->mSaveGeometry.height () / static_cast <float> (priv->mWindow->windowGeometry ().height ());

    r.setX (pointerX - ((pointerX - priv->mWindow->windowGeometry ().x ()) * xScale));
    r.setY (pointerY - ((pointerY - priv->mWindow->windowGeometry ().y ()) * yScale));

    if (target () == getDrawWindowGeometry ())
    {
	setTarget (r);
	setAnimation (AnimatableRectInterface::Type::In);
    }
    else
	setAnimation (AnimatableRectInterface::Type::Out);
}

TileLayerMiscInterface *
Tile::layer ()
{
    return priv->mParent;
}

void
Tile::setGetSnapGeometryFunc (const SnapRemovableInterface::GetGeometryFunc &f)
{
    priv->mSnapRemovable->setGetSnapGeometryFunc (f);
}

void
Tile::damage ()
{
    if (priv->mPaintable)
	priv->mPaintable->damage ();
}

TileAssignableInterface *
Tile::assigned ()
{
    return priv->mAssignable;
}


