/*
 * Compiz Tiles Plugin
 *
 * tiles.h
 *
 * Copyright (c) 2011 Sam Spilsbury <smspillaz@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#ifndef _TILES_TILELAYER_H
#define _TILES_TILELAYER_H

#include <boost/shared_ptr.hpp>
#include <boost/make_shared.hpp>

#include <core/core.h>
#include <opengl/opengl.h>

#include <core/rect.h>
#include <core/size.h>
#include <core/point.h>

#include "tile.h"
#include "grabtracker.h"
#include "mousehotspotgenerator.h"
#include "tilelayerlock.h"

class PrivateTileLayer;

class ResizableInterface;

class TileAssignableInterface
{
    public:

	virtual ~TileAssignableInterface () {}

	virtual bool assignTile (TileInterface::Ptr) = 0;
	virtual void releaseTile () = 0;
	virtual const TileInterface::Ptr & tile () const = 0;
};

class TileLayerMiscInterface
{
    public:

	typedef boost::shared_ptr <TileLayerMiscInterface> Ptr;

	virtual ~TileLayerMiscInterface () {}

	typedef unsigned int WindowState;

	static const unsigned int HorizontallyMaximized = (1 << 0);
	static const unsigned int VerticallyMaximized = (1 << 1);

	virtual void handleStateLoss (TileAssignableInterface *ta,
				      WindowState             lostState) = 0;

	virtual void handleWindowResize (ResizableInterface *tw,
					 int x,
					 int y,
					 int width,
					 int height) = 0;
	virtual const std::vector <TileInterface::Ptr> & tiles () = 0;
	virtual void paintOutline (const GLMatrix &) = 0;
	virtual bool animate (unsigned int ms) = 0;
	virtual void damage () = 0;

	virtual void setSizeBounds (const CompSize &min, const CompSize &max) = 0;

	virtual bool addWindow (ResizableInterface *tw,
				TileAssignableInterface *ta,
				DrawableWindowInterface *dwi,
				const CompRect &r,
				MaximizationStateApplicationInterface *max,
				const boost::shared_ptr <AnimatableRectInterface> &animatable,
				const boost::shared_ptr <DroppableInterface> &droppable,
				const boost::shared_ptr <DraggableInterface> &draggable,
				const boost::shared_ptr <SnapRemovableInterface> &snapRemovable,
				const boost::shared_ptr <ResizeAllocationInterface> &resizeAllocator) = 0;
	virtual void removeWindow (TileAssignableInterface *) = 0;
};

class TileLayerTrackingGrabsInterface :
    public TileLayerMiscInterface,
    public GrabTrackingInterface
{
    public:

	virtual ~TileLayerTrackingGrabsInterface () {}
	typedef boost::shared_ptr <TileLayerTrackingGrabsInterface> Ptr;
};

class MaximizationStateApplicationInterface;

class TileLayer :
    public MouseHotspotGeneratorInterface,
    public TileLayerLockInterface,
    public TileLayerTrackingGrabsInterface
{
    public:

	typedef tile::impl::GetWorkAreaFunc GetWorkAreaFunc;

	TileLayer (const CompRect &output, const GetWorkAreaFunc &, GrabTrackingInterface *);
	virtual ~TileLayer ();

	const std::vector<TileInterface::Ptr> & tiles ();

	void handleStateLoss (TileAssignableInterface             *ta,
			      TileLayerMiscInterface::WindowState lostState);

	void handleWindowResize (ResizableInterface *tw, int x, int y, int width, int height);
	bool addGrabbedWindow (TileAssignableInterface *,
			       AnimatableWindowInterface *,
			       WindowMonitorInterface *,
			       DrawableWindowInterface *,
			       DropCommitInterface *,
			       const tile::impl::GetWorkAreaFunc &,
			       const CompPoint &);
	void dragGrabbedWindow (ResizableInterface *,
				TileAssignableInterface *,
				WindowMonitorInterface *,
				DrawableWindowInterface *,
				MaximizationStateApplicationInterface *max,
				const boost::shared_ptr <AnimatableRectInterface> &animatable,
				const boost::shared_ptr <DragAndDropInterface> &dnd,
				const boost::shared_ptr <ResizeAllocationInterface> &resizeAllocator,
				const CompPoint &p);
	void removeGrabbedWindow (DropCommitInterface *);

    	bool addWindow (ResizableInterface *tw,
			TileAssignableInterface *ta,
			DrawableWindowInterface *dwi,
			const CompRect &r,
			MaximizationStateApplicationInterface *max,
			const boost::shared_ptr <AnimatableRectInterface> &animatable,
			const boost::shared_ptr <DroppableInterface> &droppable,
			const boost::shared_ptr <DraggableInterface> &draggable,
			const boost::shared_ptr <SnapRemovableInterface> &snapRemovable,
			const boost::shared_ptr <ResizeAllocationInterface> &resizeAllocator);
	void removeWindow (TileAssignableInterface *);
	void paintOutline (const GLMatrix &);

	bool animate (unsigned int ms);
	void damage ();

	const CompRect & dragZoneForPoint (const CompPoint &) const;

	void setSizeBounds (const CompSize &min, const CompSize &max);

	typedef boost::shared_ptr <TileLayer> Ptr;

	CompRect rectToFillForHotspot (const CompPoint &r);

    protected:

	bool lockTiles ();
	bool tilesToRegion (std::vector <TileInterface::Ptr> tiles,
			    CompRegion tilesRegion,
			    CompRegion nextRegion,
			    const AnimatableRectInterface::Type & type = AnimatableRectInterface::Type::In);

    private:

	PrivateTileLayer *priv;
};

/* Unity Tile Layer */
class UnityTileLayer :
    public TileLayerTrackingGrabsInterface,
    public TileLayerLockInterface
{
    public:

	UnityTileLayer (GrabTrackingInterface *grabTracker) :
	    mGrabTracker (grabTracker),
	    mDraggingInterface (NULL)
	{
	}

	void handleStateLoss (TileAssignableInterface             *ta,
			      TileLayerMiscInterface::WindowState lostState);

	bool lockTiles ();

	bool addGrabbedWindow (TileAssignableInterface *assignable,
			       AnimatableWindowInterface *animatable,
			       WindowMonitorInterface *windowMonitor,
			       DrawableWindowInterface *drawable,
			       DropCommitInterface *dropCommit,
			       const tile::impl::GetWorkAreaFunc &getWorkArea,
			       const CompPoint                   &dragPointer);
	void removeGrabbedWindow (DropCommitInterface *dropCommit);
	void dragGrabbedWindow (ResizableInterface *resizable,
				TileAssignableInterface *assignable,
				WindowMonitorInterface *windowMonitor,
				DrawableWindowInterface *drawable,
				MaximizationStateApplicationInterface *max,
				const boost::shared_ptr <AnimatableRectInterface> &animatable,
				const boost::shared_ptr <DragAndDropInterface> &dnd,
				const boost::shared_ptr <ResizeAllocationInterface> &resizeAllocator,
				const CompPoint &dragPoint);

	void handleWindowResize (ResizableInterface *tw,
				 int x,
				 int y,
				 int width,
				 int height);
	const std::vector <TileInterface::Ptr> & tiles ();
	void paintOutline (const GLMatrix &);
	bool animate (unsigned int ms);
	void damage ();

	void setSizeBounds (const CompSize &min, const CompSize &max);

	bool addWindow (ResizableInterface *tw,
			TileAssignableInterface *ta,
			DrawableWindowInterface *dwi,
			const CompRect &r,
			MaximizationStateApplicationInterface *max,
			const boost::shared_ptr <AnimatableRectInterface> &animatable,
			const boost::shared_ptr <DroppableInterface> &droppable,
			const boost::shared_ptr <DraggableInterface> &draggable,
			const boost::shared_ptr <SnapRemovableInterface> &snapRemovable,
			const boost::shared_ptr <ResizeAllocationInterface> &resizeAllocator);
	void removeWindow (TileAssignableInterface *);

    private:

	GrabTrackingInterface *mGrabTracker;
	std::vector <TileInterface::Ptr> mTiles;
	CompPoint                        mLastDragPoint;
	ResizableInterface               *mDraggingInterface;
};


#endif
