/*
 * Compiz Tiles Plugin
 *
 * tiles.cpp
 *
 * Copyright (c) 2011 Sam Spilsbury <smspillaz@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include <unordered_map>
#include <hash_fun.h>

#include <core/region.h>
#include "grabtracker.h"
#include "monitor.h"

#include "tilelayer.h"
#include "tile.h"

namespace RectMap
{
    class Hash {
	public:
	bool operator() (const CompRect &r) const
	{
	    std::hash <int> h;

	    return h (r.area ());
	}
    };

    class Predicate {
	public:
	bool operator() (const CompRect &a, const CompRect &b) const
	{
	    return (a == b);
	}
    };

    typedef std::unordered_map<CompRect, CompRect, Hash, Predicate> Map;
}

class PrivateTileLayer
{
    public:

    bool pushSpace (const CompRect &r);
    void popSpaces ();

    std::vector<TileInterface::Ptr> mTiles;
    CompRegion             mDropZones;
    CompRect               mOutput;
    TileLayer::GetWorkAreaFunc mGetWorkAreaFunc;
    RectMap::Map           mPushedSpaces;
    CompPoint              mLastPointer;
    CompRect		   mLastDropZone;

    CompSize           mMinGeometry;
    CompSize           mMaxGeometry;

    GrabTrackingInterface *mMasterGrabTracker;
};

const std::vector<TileInterface::Ptr> &
TileLayer::tiles ()
{
    return priv->mTiles;
}

void
TileLayer::handleStateLoss (TileAssignableInterface *ta,
			    TileLayerMiscInterface::WindowState lostState)
{
}

void
TileLayer::handleWindowResize (ResizableInterface *tw, int x, int y, int width, int height)
{
    std::vector <TileInterface::Ptr>::iterator tpit = priv->mTiles.begin ();

    for (; tpit != priv->mTiles.end (); tpit++)
    {
	if ((*tpit)->window () == tw)
	{
	    break;
	}
    }

    if (tpit != priv->mTiles.end ())
    {
	std::vector <TileInterface::Ptr> edgeTiles = { *tpit };
	CompRect   wg (x, y, width, height);
	CompRegion reg = CompRegion (tw->toFrameGeometry (wg));
	CompRect   r = reg.boundingRect ();
	CompRegion old = edgeTiles.front ()->getGrabGeometry ();

	/* If this tile shares an edge with other tiles
	 * and that edge would not be shared after the resize
	 * operation occurrs, resize the other tiles too
	 * such that the edge will remain shared and add
	 * them to the "reserved" region */

	for (TileInterface::Ptr stp : priv->mTiles)
	{
	    /* Skip currently resized tile */
	    if (edgeTiles.front () == stp)
		continue;

	    /* Left edge */
	    if (stp->getDrawWindowGeometry ().left () == edgeTiles.front ()->getDrawWindowGeometry ().left () &&
		stp->getDrawWindowGeometry ().left () != r.left ())
	    {
		CompRect nr = stp->getDrawWindowGeometry ();
		CompRegion tOld = stp->getDrawWindowGeometry ();

		nr.setLeft (r.left ());

		stp->setAllocation (nr);
		stp->setTarget (nr);
		stp->setAnimation (AnimatableRectInterface::Type::Complete);

		old += tOld;
		reg += nr;
		edgeTiles.push_back (stp);
	    }
	    /* Right edge */
	    else if (stp->getDrawWindowGeometry ().right () == edgeTiles.front ()->getDrawWindowGeometry ().right () &&
		     stp->getDrawWindowGeometry ().right () != r.right ())
	    {
		CompRect nr = stp->getDrawWindowGeometry ();
		CompRegion tOld = stp->getDrawWindowGeometry ();

		nr.setRight (r.right ());

		stp->setAllocation (nr);
		stp->setTarget (nr);
		stp->setAnimation (AnimatableRectInterface::Type::Complete);

		old += tOld;
		reg += nr;
		edgeTiles.push_back (stp);
	    }
	    /* Top edge */
	    else if (stp->getDrawWindowGeometry ().top () == edgeTiles.front ()->getDrawWindowGeometry ().top () &&
		     stp->getDrawWindowGeometry ().top () != r.top ())
	    {
		CompRect nr = stp->getDrawWindowGeometry ();
		CompRegion tOld = stp->getDrawWindowGeometry ();

		nr.setTop (r.top ());

		stp->setAllocation (nr);
		stp->setTarget (nr);
		stp->setAnimation (AnimatableRectInterface::Type::Complete);

		old += tOld;
		reg += nr;
		edgeTiles.push_back (stp);
	    }
	    /* Bottom edge */
	    else if (stp->getDrawWindowGeometry ().bottom () == edgeTiles.front ()->getDrawWindowGeometry ().bottom () &&
		     stp->getDrawWindowGeometry ().bottom () != r.bottom ())
	    {
		CompRect nr = stp->getDrawWindowGeometry ();
		CompRegion tOld = stp->getDrawWindowGeometry ();

		nr.setBottom (r.bottom ());

		stp->setAllocation (nr);
		stp->setTarget (nr);
		stp->setAnimation (AnimatableRectInterface::Type::Complete);

		old += tOld;
		reg += nr;
		edgeTiles.push_back (stp);
	    }
	}

	/* Allocate a new space for this window
	 * and change the target size of other windows */
	edgeTiles.front ()->setAllocation (r);
	edgeTiles.front ()->setTarget (r);
	edgeTiles.front ()->setAnimation (AnimatableRectInterface::Type::Complete);

	r = reg.boundingRect ();

	CompRegion insertRegion = reg;
	CompRegion intersectingTilesRegion;
	std::vector <TileInterface::Ptr> intersectingTiles;

	insertRegion += CompRegion (r.x2 (), r.y (), priv->mGetWorkAreaFunc ().width () - r.x2 (), r.height ());
	insertRegion += CompRegion (r.x (), r.y2 (), r.width (), priv->mGetWorkAreaFunc ().height () - r.y2 ());
	insertRegion += CompRegion (priv->mGetWorkAreaFunc ().x (), r.y (), r.x () - priv->mGetWorkAreaFunc ().x (), r.height ());
	insertRegion += CompRegion (r.x (), priv->mGetWorkAreaFunc ().y (), r.width (), r.y () - priv->mGetWorkAreaFunc ().y ());

	CompRect re = insertRegion.boundingRect ();

	/* Now reallocate the other tiles that intersect this one */
	for (TileInterface::Ptr tp : priv->mTiles)
	{
	    CompRegion intersection;
	    bool isEdgeTile = false;

	    for (TileInterface::Ptr setp : edgeTiles)
	    {
		if (tp == setp)
		{
		    intersectingTilesRegion += old;
		    isEdgeTile = true;
		    break;
		}
	    }

	    if (isEdgeTile)
		continue;

	    intersection = (CompRegion (tp->getDrawWindowGeometry ()).intersected (insertRegion));

	    if (intersection == CompRegion (tp->getDrawWindowGeometry ()))
	    {
		intersectingTiles.push_back (tp);
		intersectingTilesRegion += tp->getDrawWindowGeometry ();
	    }
	}

	re = insertRegion.boundingRect ();

	CompRegion allowedRegion = intersectingTilesRegion - r;
	intersectingTilesRegion -= old;

	re = allowedRegion.boundingRect ();

	if (!tilesToRegion (intersectingTiles, intersectingTilesRegion, allowedRegion, AnimatableRectInterface::Type::Complete))
	    return;
    }
}

const unsigned int drop_zone_area = 20;


bool
TileLayer::addGrabbedWindow (TileAssignableInterface *assignable,
			     AnimatableWindowInterface *animatable,
			     WindowMonitorInterface *wm,
			     DrawableWindowInterface *dwi,
			     DropCommitInterface *dropCommit,
			     const tile::impl::GetWorkAreaFunc &f,
			     const CompPoint                   &dragPoint)
{
    if (!priv->mMasterGrabTracker->addGrabbedWindow (assignable,
						     animatable,
						     wm,
						     dwi,
						     dropCommit,
						     f,
						     dragPoint))
	return false;

    priv->mPushedSpaces.clear ();

    /* Recalculate drop zones
    if (tiles ().empty ())
    {
	priv->mDropZones = CompRegion (*priv->mOutput) -
		CompRegion (priv->mOutput.x () + drop_zone_area, priv->mOutput.y () + drop_zone_area,
			    priv->mOutput.width () - drop_zone_area * 2, priv->mOutput.height () - drop_zone_area * 2);
    }
    else
    {
	priv->mDropZones = CompRegion (*priv->mOutput);
	for (const TileInterface::Ptr &tp : tiles ())
	{
	    priv->mDropZones -= CompRegion (tp->allocation ().x () + drop_zone_area, tp->allocation ().y () + drop_zone_area,
				     tp->allocation ().width () - drop_zone_area * 2, tp->allocation ().height () - drop_zone_area * 2);
	}
    } */

    priv->mDropZones = CompRegion (priv->mOutput) -
	    CompRegion (priv->mOutput.x () + drop_zone_area, priv->mOutput.y () + drop_zone_area,
			priv->mOutput.width () - drop_zone_area * 2, priv->mOutput.height () - drop_zone_area * 2);

    priv->mLastPointer = dragPoint;

    return true;
}

void
TileLayer::removeGrabbedWindow (DropCommitInterface *dropCommit)
{
    priv->mMasterGrabTracker->removeGrabbedWindow (dropCommit);
}

void
TileLayer::setSizeBounds (const CompSize &minBounds,
			  const CompSize &maxBounds)
{
    priv->mMinGeometry = minBounds;
    priv->mMaxGeometry = maxBounds;
}

namespace
{

void
calculateRects (int left, int top,
		const std::list <int> &bottomEdges,
		const std::list <int> &rightEdges,
		std::vector <CompRect::vector> &rects)
{
    int initialLeft = left;
    std::vector <CompRect::vector>::iterator columnIt = rects.begin ();

    for (const int &bottom : bottomEdges)
    {
	CompRect::vector::iterator rowIt = (*columnIt).begin ();

	for (const int &right : rightEdges)
	{
	    (*rowIt).setGeometry (left, top,
				  right - left, bottom - top);
	    rowIt++;
	    left = right;
	}

	columnIt++;
	top = bottom;
	left = initialLeft;
    }
}

unsigned int
calculateRowIntersections (const std::vector <TileInterface::Ptr> &tiles,
			   const std::vector <CompRect::vector> &rects,
			   unsigned int rowNum)
{
    CompRegion rowRegion;
    unsigned int nIntersections = 0;

    for (const CompRect &r : rects[rowNum])
    	rowRegion += r;

    for (const TileInterface::Ptr &tp : tiles)
    	if (rowRegion.intersects (tp->allocation ()))
    	    nIntersections++;

    return nIntersections;
}

unsigned int
calculateColumnIntersections (const std::vector <TileInterface::Ptr> &tiles,
			      const std::vector <CompRect::vector> &rects,
			      unsigned int columnNum)
{
    CompRegion rowRegion;
    unsigned int nIntersections = 0;

    for (const CompRect::vector &rv : rects)
    	rowRegion += rv[columnNum];

    for (const TileInterface::Ptr &tp : tiles)
    	if (rowRegion.intersects (tp->allocation ()))
    	    nIntersections++;

    return nIntersections;
}

}

bool
PrivateTileLayer::pushSpace (const CompRect &r)
{
    CompRect available = mGetWorkAreaFunc ();
    unsigned int nTiles = 0;
    unsigned int maxWidth = mGetWorkAreaFunc ().width () / mMinGeometry.width ();
    unsigned int minWidth = mGetWorkAreaFunc ().width () / mMaxGeometry.width ();

    unsigned int maxHeight = mGetWorkAreaFunc ().height () / mMinGeometry.height ();
    unsigned int minHeight = mGetWorkAreaFunc ().height () / mMaxGeometry.height ();

    if (mPushedSpaces.find (r) != mPushedSpaces.end ())
	return true;

    /* Create lists of all of the bottom and left edges that already
     * exist in this tile layer .. sort them lowest ot highest */
    std::list <int> bottomEdges;
    std::list <int> rightEdges;

    CompRegion tilesRegion;

    for (const TileInterface::Ptr &tp : mTiles)
    {
    	bottomEdges.push_back (tp->allocation ().bottom ());
    	rightEdges.push_back (tp->allocation ().right ());
    	tilesRegion += tp->allocation ();
    }

    CompRegion freeRegion (mGetWorkAreaFunc ());
    freeRegion -= tilesRegion;

    std::unique (bottomEdges.begin (), bottomEdges.end ());
    std::unique (rightEdges.begin (), rightEdges.end ());

    std::vector <CompRect::vector> rects (bottomEdges.size ());

    for (CompRect::vector &v : rects)
    	v.resize (rightEdges.size ());

    calculateRects (mGetWorkAreaFunc ().left (),
		    mGetWorkAreaFunc ().top (),
		    bottomEdges, rightEdges, rects);

    /* Calculate what this space expands to :
     * If this stip is longer than it is wider,
     * expand it horizontally, otherwise
     * vertically */

    if (r.height () > r.width ())
    {
	available.setHeight (r.height ());
	available.setY (r.y ());

	/* The way the region calculation works out
	 * means that there might be some extra space
	 * so work out if there are any other tiles partially
	 * overlapping workArea.height () * r.width ()
	 * and then add the empty spaces */
	CompRegion bar = CompRegion (r.x (), mGetWorkAreaFunc ().y (), r.width (), mGetWorkAreaFunc ().height ());
	CompRegion origBar = bar;

	/* Count the smallest number of tiles rightwards ..
	 * this will be used in determining the scale factor
	 * for the newly inserted tile */
	for (const TileInterface::Ptr &tp : mTiles)
	    bar -= tp->allocation ();

	nTiles = 0;
	for (unsigned int i  = 0; i < rightEdges.size (); ++i)
	{
	    nTiles = std::max (nTiles, calculateRowIntersections (mTiles, rects, i));
	}

	if (bar.isEmpty ())
	{
	    available.setY (origBar.boundingRect ().y ());
	    available.setHeight (origBar.boundingRect ().height ());
	    /* Divide by number of tiles + 1, this will be
	     * the space we need to occupy */
	    available.setWidth (available.width () / (nTiles + 1));

	    unsigned int width = std::min (maxWidth, static_cast <unsigned int> (available.width ()));
	    int rectX2 = r.x () + width;

	    /* Prefer going left to right unless we'd go offscreen that way */
	    if (rectX2 > mOutput.width ())
		available.setX (r.x2 () - width);
	    else
		available.setX (r.x ());
	}
	else if (!freeRegion.isEmpty ())
	{
	    /* The drag area occupies some free space,
	     * so find the rest of the free space and
	     * prefer using that */
	    CompRegion expandedBar (mGetWorkAreaFunc ().x (),
				    bar.boundingRect ().y (), mGetWorkAreaFunc ().width (), bar.boundingRect ().height ());

	    expandedBar &= freeRegion;

	    if (!expandedBar.isEmpty ())
	    {
		CompRect::vector rects = expandedBar.rects ();
		CompRect *thinnest = &(rects[0]);

		for (CompRect &r : rects)
		    if (thinnest->width () > r.width ())
			thinnest = &r;

		available = *thinnest;
	    }
	    else
		return false;
	}

	available.setWidth (std::min (maxWidth, static_cast <unsigned int> (available.width ())));
	available.setHeight (std::min (maxHeight, static_cast <unsigned int> (available.height ())));

	if (static_cast <unsigned int> (available.width ()) < minWidth)
	    return false;
    }
    else
    {
	available.setWidth (r.width ());
	available.setX (r.x ());

	/* The way the region calculation works out
	 * means that there might be some extra space
	 * so work out if there are any other tiles partially
	 * overlapping workArea.height () * r.width ()
	 * and then add the empty spaces */
	CompRegion bar = CompRegion (mGetWorkAreaFunc ().x (), r.y (), mGetWorkAreaFunc ().width (), r.height ());
	CompRegion origBar = bar;

	/* Count the smallest number of tiles rightwards ..
	 * this will be used in determining the scale factor
	 * for the newly inserted tile */
	for (const TileInterface::Ptr &tp : mTiles)
	    bar -= tp->allocation ();

	nTiles = 0;
	for (unsigned int i  = 0; i < bottomEdges.size (); ++i)
	{
	    nTiles = std::max (nTiles, calculateColumnIntersections (mTiles, rects, i));
	}

	if (bar.isEmpty ())
	{
	    available.setX (origBar.boundingRect ().x ());
	    available.setWidth (origBar.boundingRect ().width ());
	    /* Divide by number of tiles + 1, this will be
	     * the space we need to occupy */
	    available.setHeight (available.height () / (nTiles + 1));

	    unsigned int height = std::min (maxHeight, static_cast <unsigned int> (available.height ()));

	    int rectY2 = r.y () + height;
	    /* Prefer going top to bottom unless we'd go offscreen that way */
	    if (rectY2 > mOutput.height ())
		available.setY (r.y2 () - height);
	    else
		available.setY (r.y ());
	}
	else if (!freeRegion.isEmpty ())
	{
	    /* The drag area occupies some free space,
	     * so find the rest of the free space and
	     * prefer using that */
	    CompRegion expandedBar (bar.boundingRect ().x (),
				    mGetWorkAreaFunc ().y (), bar.boundingRect ().width (), mGetWorkAreaFunc ().height ());

	    expandedBar &= freeRegion;

	    if (!expandedBar.isEmpty ())
	    {
		CompRect::vector rects = expandedBar.rects ();
		CompRect *shortest = &(rects[0]);

		for (CompRect &r : expandedBar.rects ())
		    if (shortest->height () > r.height ())
			shortest = &r;

		available = *shortest;
	    }
	    else
		return false;
	}

	available.setWidth (std::min (maxWidth, static_cast <unsigned int> (available.width ())));
	available.setHeight (std::min (maxHeight, static_cast <unsigned int> (available.height ())));

	if (static_cast <unsigned int> (available.height ()) < minHeight)
	    return false;
    }

    mPushedSpaces[r] = available;

    return true;
}

void
PrivateTileLayer::popSpaces ()
{
    mPushedSpaces.clear ();
}

bool
TileLayer::lockTiles ()
{
    for (const TileInterface::Ptr &tp : priv->mTiles)
	tp->lockTargetSize ();

    return true;
}

void
TileLayer::dragGrabbedWindow (ResizableInterface *tw,
			      TileAssignableInterface *ta,
			      WindowMonitorInterface *wm,
			      DrawableWindowInterface *dwi,
			      MaximizationStateApplicationInterface *max,
			      const boost::shared_ptr <AnimatableRectInterface> &animatable,
			      const boost::shared_ptr <DragAndDropInterface> &dnd,
			      const boost::shared_ptr <ResizeAllocationInterface> &resizeAllocator,
			      const CompPoint &p)
{
    std::vector <TileInterface::Ptr>::iterator tpit = priv->mTiles.begin ();

    priv->mMasterGrabTracker->dragGrabbedWindow (tw, ta, wm, dwi, max, animatable,
						 dnd,
						 resizeAllocator, p);

    for (; tpit != priv->mTiles.end (); tpit++)
    {
	if ((*tpit)->window () == tw)
	    break;
    }

    if (tpit == priv->mTiles.end ())
    {
	CompRect r = rectToFillForHotspot (p);

	if (!r.isEmpty ())
	    addWindow (tw, ta, dwi, r, max, animatable,
		       boost::shared_static_cast <DroppableInterface> (dnd),
		       boost::shared_static_cast <DraggableInterface> (dnd),
		       boost::shared_static_cast <SnapRemovableInterface> (dnd), resizeAllocator);
	else
	    removeWindow (ta);
    }
    else if ((*tpit)->layer ())
    {
	CompPoint dp = p - priv->mLastPointer;

	(*tpit)->setPotentialDropZone (dragZoneForPoint (priv->mLastPointer));
	(*tpit)->applyDrag (dp);

	if ((*tpit)->shouldSnapOut ())
	    removeWindow (ta);

    }

    priv->mLastPointer = p;
}

CompRect
TileLayer::rectToFillForHotspot (const CompPoint &p)
{
    CompRect *rect = NULL;

    for (CompRect r : priv->mDropZones.rects ())
    {
	if (CompRegion (r).contains (p))
	{
	    CompRegion clip = priv->mGetWorkAreaFunc ();
#if 0
	    if (!priv->mTiles.size ())
	    {
		if (r.width () > r.height ())
		{
		    /* Special case window maximization
		     * at the top */
		    if (r.top () == clip.boundingRect ().top ())
			setSizeBounds (CompSize (1, 1), CompSize (1, 1));
		}
		else if (r.height () > r.width ())
		{
		    /* Special case left and right tiles */
		    if (r.left () == clip.boundingRect ().left ())
			setSizeBounds (CompSize (2, 1), CompSize (2, 1));
		    else if (r.right () == clip.boundingRect ().right ())
			setSizeBounds (CompSize (2, 1), CompSize (2, 1));
		}
	    }
#endif
	    /* expand r to next tile space */
	    if (priv->pushSpace (r))
		rect = &priv->mPushedSpaces[r];
	}
    }

    if (rect)
	return *rect;
    else
    {
	priv->popSpaces ();
	return CompRect ();
    }
}

bool
TileLayer::tilesToRegion (std::vector <TileInterface::Ptr> tiles,
			  CompRegion tilesRegion,
			  CompRegion nextRegion,
			  const AnimatableRectInterface::Type &type)
{
    for (TileInterface::Ptr tp : tiles)
    {
	float xScale = nextRegion.boundingRect ().width () / static_cast <float> (tilesRegion.boundingRect ().width ());
	float yScale = nextRegion.boundingRect ().height () / static_cast <float> (tilesRegion.boundingRect ().height ());

	CompPoint pos = tp->getDrawWindowGeometry ().pos () - tilesRegion.boundingRect ().pos ();

	float width = tp->getDrawWindowGeometry ().width () * xScale;
	float height = tp->getDrawWindowGeometry ().height () * yScale;
	float x = pos.x () * xScale;
	float y = pos.y () * yScale;

	if (xScale >= 1.0f)
	{
	    width = ceil (width);
	    x = ceil (x);
	}
	else
	{
	    width = floor (width);
	    x = floor (x);
	}

	if (yScale >= 1.0f)
	{
	    height = ceil (height);
	    y = ceil (y);
	}
	else
	{
	    height = floor (height);
	    y = floor (y);
	}

	CompRect nextPosition;

	/* Scale the new positionfs */
	pos.setX (x);
	pos.setY (y);

	pos += nextRegion.boundingRect ().pos ();

	/* Now scale the sizes */
	nextPosition.setWidth (width);
	nextPosition.setHeight (height);

	nextPosition.setPos (pos);

	tp->setAllocation (nextPosition);
	tp->setTarget (nextPosition);

	tp->setAnimation (type);
    }

    return true;
}

bool
TileLayer::addWindow (ResizableInterface *tw,
		      TileAssignableInterface *ta,
		      DrawableWindowInterface *dwi,
		      const CompRect &r,
		      MaximizationStateApplicationInterface *max,
		      const AnimatableRectInterface::Ptr &animatable,
		      const DroppableInterface::Ptr &droppable,
		      const DraggableInterface::Ptr &draggable,
		      const SnapRemovableInterface::Ptr &snapRemovable,
		      const ResizeAllocationInterface::Ptr &resizeAllocator)
{
    boost::function <bool ()> dropAction (boost::bind (&TileLayer::lockTiles, this));

    TileInterface::Ptr t = TileInterface::Ptr (new Tile (this,
							 animatable,
							 dwi,
							 droppable,
							 draggable,
							 snapRemovable,
							 resizeAllocator,
							 this,
							 max,
							 dropAction));
    snapRemovable->setGetSnapGeometryFunc (boost::bind (&TileInterface::allocation, t.get ()));
    std::vector <TileInterface::Ptr> intersectingTiles;
    CompRegion              intersectingTilesRegion;
    CompRegion              insertRegion (r);

    /* Expand like this:
    *             -------------
     *            |           |
     *            |           |
     *            |           |
     *            |           |
     *            |           |
     * ------------------------------------
     * |          |           |           |
     * |          |           |           |
     * |          |           |           |
     * |          |           |           |
     * ------------------------------------
     *            |           |
     *            |           |
     *            |           |
     *            |           |
     *            |           |
     *            -------------
     */

    insertRegion += CompRegion (r.x2 (), r.y (), priv->mGetWorkAreaFunc ().width () - r.x2 (), r.height ());
    insertRegion += CompRegion (r.x (), r.y2 (), r.width (), priv->mGetWorkAreaFunc ().height () - r.y2 ());
    insertRegion += CompRegion (priv->mGetWorkAreaFunc ().x (), r.y (), r.x () - priv->mGetWorkAreaFunc ().x (), r.height ());
    insertRegion += CompRegion (r.x (), priv->mGetWorkAreaFunc ().y (), r.width (), r.y () - priv->mGetWorkAreaFunc ().y ());

    if (!t->allocateWindow (tw, ta))
	return false;
    else
	ta->assignTile (t);

    /* Allocate a new space for this window
     * and change the target size of other windows */
    t->setAllocation (r);
    t->setTarget (r);

    t->setAnimation (AnimatableRectInterface::Type::In);

    /* Now reallocate the other tiles that intersect this one */
    for (TileInterface::Ptr tp : priv->mTiles)
    {
	if ((CompRegion (tp->allocation ()) - insertRegion).isEmpty ())
	{
	    intersectingTiles.push_back (tp);
	    intersectingTilesRegion += tp->allocation ();
	}
    }

    CompRegion allowedRegion = intersectingTilesRegion - r;

    if (!tilesToRegion (intersectingTiles, intersectingTilesRegion, allowedRegion))
	return false;

    priv->mTiles.push_back (t);

    return true;
}

void
TileLayer::removeWindow (TileAssignableInterface *ta)
{
    std::vector <TileInterface::Ptr>::iterator tpit = priv->mTiles.begin ();
    std::vector <TileInterface::Ptr> resizeTiles;
    CompRegion              tilesRegion;

    TileInterface::Ptr tip = ta->tile ();

    tpit = std::find (priv->mTiles.begin (), priv->mTiles.end (), ta->tile ());

    if (tpit == priv->mTiles.end ())
    	return;


    (*tpit)->disassociate ();
    priv->mTiles.erase (tpit);

    /* Now calculate the empty space and find any tiles contained
     * in that space and expand them evenly to fill the empty space
     * created by the newly removed tile */

    CompRegion reg = priv->mGetWorkAreaFunc ();

    for (TileInterface::Ptr tp : priv->mTiles)
    	reg -= tp->allocation ();

    CompRegion insertRegion = reg;
    CompRect   r = reg.boundingRect ();

    /* Expand the rect towards screen edges (as illustrated above) */
    insertRegion += CompRegion (r.x2 (), r.y (), priv->mGetWorkAreaFunc ().width () - r.x2 (), r.height ());
    insertRegion += CompRegion (r.x (), r.y2 (), r.width (), priv->mGetWorkAreaFunc ().height () - r.y2 ());
    insertRegion += CompRegion (priv->mGetWorkAreaFunc ().x (), r.y (), r.x () - priv->mGetWorkAreaFunc ().x (), r.height ());
    insertRegion += CompRegion (r.x (), priv->mGetWorkAreaFunc ().y (), r.width (), r.y () - priv->mGetWorkAreaFunc ().y ());

    /* Now find any windows which are directly contained by this
     * cross-shaped region */
    for (TileInterface::Ptr tp : priv->mTiles)
    {
    	CompRegion rem (tp->allocation ());
    	CompRegion other = insertRegion.intersected (rem);

	if (rem == other)
	{
	    resizeTiles.push_back (tp);
	    tilesRegion += tp->allocation ();
	}
    }

    reg += tilesRegion;

    tilesToRegion (resizeTiles, tilesRegion, reg);
}

bool
TileLayer::animate (unsigned int ms)
{
    bool continueAnimation = false;

    for (const TileInterface::Ptr &p : tiles ())
	continueAnimation |= p->animate (ms);

    if (continueAnimation)
    {
	/* Check the layer animation too */
	// ??
    }

    return continueAnimation;
}

void
TileLayer::paintOutline(const GLMatrix &m)
{
    return;
}



void
TileLayer::damage ()
{
    for (const TileInterface::Ptr &p : tiles ())
	p->damage ();
}

const CompRect &
TileLayer::dragZoneForPoint (const CompPoint &p) const
{
    for (const CompRect &r : priv->mDropZones.rects ())
    {
	if (r.contains (p))
	{
	    priv->mLastDropZone = r;
	    return priv->mLastDropZone;
	}
    }

    return priv->mLastDropZone;
}

TileLayer::TileLayer (const CompRect &output,
		      const GetWorkAreaFunc &getWorkArea,
		      GrabTrackingInterface *masterTracker) :
    priv (new PrivateTileLayer)
{
    priv->mOutput = output;
    priv->mGetWorkAreaFunc = getWorkArea;
    priv->mMinGeometry = CompSize (1, 1);
    priv->mMaxGeometry = CompSize (std::numeric_limits <int>::max (),
				   std::numeric_limits <int>::max ());
    priv->mMasterGrabTracker = masterTracker;
}

TileLayer::~TileLayer ()
{
    delete priv;
}

/* Unity Tile Layer impl */

void
UnityTileLayer::paintOutline(const GLMatrix &mat)
{
}

bool
UnityTileLayer::animate(unsigned int ms)
{
    bool continueAnimation = false;

    for (const TileInterface::Ptr &tp : mTiles)
	continueAnimation |= tp->animate (ms);

    return continueAnimation;
}

void
UnityTileLayer::damage()
{
    for (const TileInterface::Ptr &tp : mTiles)
	tp->damage();
}

void
UnityTileLayer::setSizeBounds(const CompSize &min, const CompSize &max)
{
}

bool
UnityTileLayer::addWindow(ResizableInterface *resizable, TileAssignableInterface *assignable, DrawableWindowInterface *drawable, const CompRect &r, MaximizationStateApplicationInterface *max, const boost::shared_ptr<AnimatableRectInterface> &animatable, const boost::shared_ptr<DroppableInterface> &droppable, const boost::shared_ptr<DraggableInterface> &draggable, const boost::shared_ptr<SnapRemovableInterface> &snapRemovable, const boost::shared_ptr<ResizeAllocationInterface> &resizeAllocator)
{
    TileInterface::Ptr tile (new Tile (this,
			     animatable,
			     drawable,
			     droppable,
			     draggable,
			     snapRemovable,
			     resizeAllocator,
			     this,
			     max));

    if (!tile->allocateWindow (resizable, assignable))
	return false;

    assignable->assignTile (tile);
    tile->setAllocation (r);
    tile->setTarget (r);
    tile->setAnimation (AnimatableRectInterface::Type::In);

    snapRemovable->setGetSnapGeometryFunc (boost::bind (&TileInterface::allocation, tile.get ()));

    mTiles.push_back (tile);

    return true;
}

void
UnityTileLayer::removeWindow (TileAssignableInterface *assignable)
{
    TileInterface::Ptr tp (assignable->tile ());

    /* Obviously bail out if this window is not even tiled */
    if (!tp)
	return;

    std::vector <TileInterface::Ptr>::iterator it (std::find (mTiles.begin (),
							      mTiles.end (),
							      tp));

    if (it != mTiles.end ())
	mTiles.erase (it);

    assignable->tile()->disassociate();
}

bool
UnityTileLayer::lockTiles ()
{
    return true;
}

bool
UnityTileLayer::addGrabbedWindow (TileAssignableInterface *assignable, AnimatableWindowInterface *animatable, WindowMonitorInterface *windowMonitor, DrawableWindowInterface *drawable, DropCommitInterface *dropCommit, const tile::impl::GetWorkAreaFunc &getWorkArea, const CompPoint &dragPointer)
{
    /* Don't track already tiled windows, we don't even
     * want the master grab tracker to know about them */
    if (std::find (mTiles.begin (),
		   mTiles.end (),
		   assignable->tile ()) != mTiles.end ())
    {
	return false;
    }

    if (!mGrabTracker->addGrabbedWindow (assignable,
					 animatable,
					 windowMonitor,
					 drawable,
					 dropCommit,
					 getWorkArea,
					 dragPointer))
	return false;

    mLastDragPoint = dragPointer;

    return true;
}

void
UnityTileLayer::dragGrabbedWindow (ResizableInterface *resizable, TileAssignableInterface *assignable, WindowMonitorInterface *windowMonitor, DrawableWindowInterface *drawable, MaximizationStateApplicationInterface *maxStateInterface, const boost::shared_ptr<AnimatableRectInterface> &animatable, const boost::shared_ptr<DragAndDropInterface> &dnd, const boost::shared_ptr<ResizeAllocationInterface> &resizeAllocator, const CompPoint &dragPoint)
{
    mGrabTracker->dragGrabbedWindow(resizable, assignable, windowMonitor, drawable, maxStateInterface, animatable, dnd, resizeAllocator, dragPoint);

    const unsigned int dropAreaSize = 15;

    const CompRect &monitorWorkArea (windowMonitor->monitorWorkArea ());
    CompRegion noDropRegion (monitorWorkArea.x1 () + dropAreaSize,
			     monitorWorkArea.y1 () + dropAreaSize,
			     monitorWorkArea.width () - dropAreaSize * 2,
			     monitorWorkArea.height () - dropAreaSize);

    CompRegion dropRegion (CompRegion (monitorWorkArea) - noDropRegion);

    TileInterface::Ptr tile (assignable->tile ());

    for (const CompRect &dropArea : dropRegion.rects ())
    {
	if (dropArea.contains (dragPoint))
	{
	    const unsigned int halfWidth = monitorWorkArea.width () / 2;

	    const CompRect &max (monitorWorkArea);
	    const CompRect left (monitorWorkArea.x (),
				 monitorWorkArea.y (),
				 halfWidth,
				 monitorWorkArea.height ());
	    const CompRect right (monitorWorkArea.x () + halfWidth,
				  monitorWorkArea.y (),
				  halfWidth,
				  monitorWorkArea.height ());

	    const CompRect *destination = NULL;

	    /* Top */
	    if (dropArea.width () > dropArea.height ())
		destination = &max;
	    else
	    {
		/* Left */
		if (dropArea.centerX () < monitorWorkArea.centerX ())
		    destination = &left;
		/* Right */
		else
		    destination = &right;
	    }

	    if (!tile)
	    {
		if (!addWindow (resizable,
				assignable,
				drawable,
				*destination,
				maxStateInterface,
				animatable,
				boost::shared_static_cast <DroppableInterface> (dnd),
				boost::shared_static_cast <DraggableInterface> (dnd),
				boost::shared_static_cast <SnapRemovableInterface> (dnd),
				resizeAllocator))
		    continue;
		else
		    tile = assignable->tile ();

	    }

	    tile->setPotentialDropZone (dropArea);
	    break;
	}
    }

    if (tile)
    {
	tile->applyDrag (dragPoint - mLastDragPoint);

	if (tile->shouldSnapOut ())
	{
	    removeWindow (assignable);
	}
    }

    mLastDragPoint = dragPoint;
    mDraggingInterface = resizable;
}

void
UnityTileLayer::removeGrabbedWindow(DropCommitInterface *dropCommit)
{
    mGrabTracker->removeGrabbedWindow(dropCommit);
}

void
UnityTileLayer::handleStateLoss (TileAssignableInterface *ta, TileLayerMiscInterface::WindowState lostState)
{
    if (lostState & TileLayerMiscInterface::VerticallyMaximized)
	removeWindow (ta);
}

void
UnityTileLayer::handleWindowResize(ResizableInterface *tw, int x, int y, int width, int height)
{
}

const std::vector <TileInterface::Ptr> &
UnityTileLayer::tiles ()
{
    return mTiles;
}

