#include <tr1/tuple>

#include <gtest/gtest.h>
#include <gmock/gmock.h>

#include "tiles-test-utils.h"

#include "mockanimatable.h"
#include "mockdrawable.h"
#include "mockassignable.h"
#include "mockresizable.h"
#include "mockmonitor.h"
#include "mockdnd.h"
#include "mockmaximization.h"
#include "mockgrabtracker.h"
#include "mockanimatedrect.h"
#include "mockresizeallocator.h"
#include "mocktile.h"
#include "animatedrects.h"
#include "tilelayer.h"

/* FIXME: Don't copy and paste value types ... fix
 * the opengl plugin */
GLVector::GLVector ()
{
    memset (v, 0, sizeof (v));
}

GLVector::GLVector (float x,
			float y,
			float z,
			float w)
{
    v[0] = x;
    v[1] = y;
    v[2] = z;
    v[3] = w;
}

float &
GLVector::operator[] (int item)
{
    return v[item];
}

float &
GLVector::operator[] (VectorCoordsEnum coord)
{
    int item = (int) coord;
    return v[item];
}

const float &
GLVector::operator[] (int item) const
{
    return v[item];
}

const float &
GLVector::operator[] (VectorCoordsEnum coord) const
{
    int item = (int) coord;
    return v[item];
}

using ::testing::Return;
using ::testing::ReturnRef;
using ::testing::_;
using ::testing::AtLeast;
using ::testing::Invoke;
using ::testing::StrictMock;
using ::testing::ValuesIn;
using ::testing::Combine;
using ::testing::WithArgs;


class TilesTileLayerTest:
    public ::testing::Test
{
};

class FakeTileAssignable :
    public TileAssignableInterface
{
    public:
	bool assignTile (TileInterface::Ptr t) { mTile = t; return true; }
	void releaseTile () { mTile.reset (); }
	const TileInterface::Ptr & tile () const;

    private:

	TileInterface::Ptr mTile;
};

const TileInterface::Ptr &
FakeTileAssignable::tile () const
{
    return mTile;
}

TEST(TilesTileLayerTest, TestAddGrabbedWindow)
{
    MockWindowMonitor mwm;
    MockGrabTracker   mgt;
    MockAssignable mta;
    MockAnimatableWindow maw;
    MockDrawable md;
    MockDropCommit mdc;
    MockResizable mr;
    CompPoint     dragPoint;

    TileLayer::GetWorkAreaFunc gwaf (boost::bind (&WindowMonitorInterface::monitorWorkArea, &mwm));

    CompRegion outputr (0, 0, 1000, 1000);

    EXPECT_CALL (mwm, monitorRegion ()).WillOnce (Return (outputr));
    EXPECT_CALL (mgt, addGrabbedWindow (&mta, &maw, &mwm, &md, &mdc, _, dragPoint));

    TileLayer tl (mwm.monitorRegion ().boundingRect (),
		  gwaf,
		  &mgt);
    tl.addGrabbedWindow (&mta, &maw, &mwm, &md, &mdc, gwaf, dragPoint);
}

TEST(TilesTileLayerTest, TestAddGrabbedWindowDragZonesSetCorrectly)
{
    MockWindowMonitor mwm;
    MockGrabTracker   mgt;
    MockAssignable mta;
    MockAnimatableWindow maw;
    MockDrawable md;
    MockDropCommit mdc;
    MockResizable mr;
    CompPoint     dragPoint;

    TileLayer::GetWorkAreaFunc gwaf (boost::bind (&WindowMonitorInterface::monitorWorkArea, &mwm));

    CompRegion outputr (0, 0, 1000, 1000);

    EXPECT_CALL (mwm, monitorRegion ()).WillOnce (Return (outputr));
    EXPECT_CALL (mgt, addGrabbedWindow (&mta, &maw, &mwm, &md, &mdc, _, dragPoint));

    TileLayer tl (mwm.monitorRegion ().boundingRect (),
		  gwaf,
		  &mgt);
    tl.addGrabbedWindow (&mta, &maw, &mwm, &md, &mdc, gwaf, dragPoint);

    EXPECT_EQ (CompRect (0, 0, 1000, 20), tl.dragZoneForPoint (CompPoint (0, 0)));
    EXPECT_EQ (CompRect (0, 0, 1000, 20), tl.dragZoneForPoint (CompPoint (1000, 0)));
    EXPECT_EQ (CompRect (0, 0, 1000, 20), tl.dragZoneForPoint (CompPoint (1000, 20)));
    EXPECT_EQ (CompRect (0, 0, 1000, 20), tl.dragZoneForPoint (CompPoint (0, 20)));

    EXPECT_EQ (CompRect (0, 20, 20, 960), tl.dragZoneForPoint (CompPoint (0, 21)));
    EXPECT_EQ (CompRect (0, 20, 20, 960), tl.dragZoneForPoint (CompPoint (20, 21)));
    EXPECT_EQ (CompRect (0, 20, 20, 960), tl.dragZoneForPoint (CompPoint (20, 980)));
    EXPECT_EQ (CompRect (0, 20, 20, 960), tl.dragZoneForPoint (CompPoint (0, 980)));

    EXPECT_EQ (CompRect (0, 980, 1000, 20), tl.dragZoneForPoint (CompPoint (0, 981)));
    EXPECT_EQ (CompRect (0, 980, 1000, 20), tl.dragZoneForPoint (CompPoint (1000, 981)));
    EXPECT_EQ (CompRect (0, 980, 1000, 20), tl.dragZoneForPoint (CompPoint (0, 1000)));
    EXPECT_EQ (CompRect (0, 980, 1000, 20), tl.dragZoneForPoint (CompPoint (1000, 981)));

    EXPECT_EQ (CompRect (980, 20, 20, 960), tl.dragZoneForPoint (CompPoint (981, 21)));
    EXPECT_EQ (CompRect (980, 20, 20, 960), tl.dragZoneForPoint (CompPoint (981, 980)));
    EXPECT_EQ (CompRect (980, 20, 20, 960), tl.dragZoneForPoint (CompPoint (1000, 980)));
    EXPECT_EQ (CompRect (980, 20, 20, 960), tl.dragZoneForPoint (CompPoint (1000, 21)));
}

TEST(TilesTileLayerTest, TestAddGrabbedWindowCorrectRectsToFill)
{
    MockWindowMonitor mwm;
    MockGrabTracker   mgt;
    MockAssignable mta;
    MockAnimatableWindow maw;
    MockDrawable md;
    MockDropCommit mdc;
    MockResizable mr;
    CompPoint     dragPoint;

    TileLayer::GetWorkAreaFunc gwaf (boost::bind (&WindowMonitorInterface::monitorWorkArea, &mwm));

    CompRegion outputr (0, 0, 1000, 1000);
    CompRect workArea (0, 0, 1000, 1000);

    EXPECT_CALL (mwm, monitorRegion ()).WillOnce (Return (outputr));
    EXPECT_CALL (mwm, monitorWorkArea ()).WillRepeatedly (Return (workArea));
    EXPECT_CALL (mgt, addGrabbedWindow (&mta, &maw, &mwm, &md, &mdc, _, dragPoint));

    TileLayer tl (mwm.monitorRegion ().boundingRect (),
		  gwaf,
		  &mgt);
    tl.addGrabbedWindow (&mta, &maw, &mwm, &md, &mdc, gwaf, dragPoint);

    EXPECT_EQ (CompRect (0, 0, 1000, 1000), tl.rectToFillForHotspot (CompPoint (0, 0)));
    EXPECT_EQ (CompRect (0, 0, 1000, 1000), tl.rectToFillForHotspot (CompPoint (0, 999)));
    EXPECT_EQ (CompRect (0, 0, 1000, 1000), tl.rectToFillForHotspot (CompPoint (999, 979)));
    EXPECT_EQ (CompRect (0, 0, 1000, 1000), tl.rectToFillForHotspot (CompPoint (0, 21)));
}

TEST(TilesTileLayerTest, TestDragWindowToDragZoneNone)
{
    MockWindowMonitor mwm;
    MockGrabTracker   mgt;
    MockAssignable mta;
    MockAnimatableWindow maw;
    MockDrawable md;
    MockDropCommit mdc;
    MockResizable mr;
    MockMaximziationStateApplier mmx;
    CompPoint     dragPoint;

    TileLayer::GetWorkAreaFunc gwaf (boost::bind (&WindowMonitorInterface::monitorWorkArea, &mwm));

    CompRegion outputr (0, 0, 1000, 1000);
    CompRect workArea (0, 0, 1000, 1000);

    EXPECT_CALL (mwm, monitorRegion ()).WillOnce (Return (outputr));
    EXPECT_CALL (mwm, monitorWorkArea ()).WillRepeatedly (Return (workArea));
    EXPECT_CALL (mgt, addGrabbedWindow (&mta, &maw, &mwm, &md, &mdc, _, dragPoint));

    TileLayer tl (mwm.monitorRegion ().boundingRect (),
		  gwaf,
		  &mgt);
    tl.addGrabbedWindow (&mta, &maw, &mwm, &md, &mdc, gwaf, dragPoint);

    CompRegion noDragZoneReg = outputr;

    std::vector <CompPoint> points =
    {
	CompPoint (0, 0),
	CompPoint (0, 999),
	CompPoint (999, 979),
	CompPoint (0, 21)
    };

    for (const CompPoint &p : points)
	noDragZoneReg - tl.dragZoneForPoint (p);

    CompPoint center (noDragZoneReg.boundingRect ().center ());

    TileInterface::Ptr nilTile (boost::make_shared <MockTile> ());

    EXPECT_CALL (mgt, dragGrabbedWindow (_, _, _, _, _, _, _, _, _));
    EXPECT_CALL (mta, tile ()).WillRepeatedly (ReturnRef (nilTile));

    boost::shared_ptr <AnimatableRectInterface> ar (boost::make_shared <MockAnimatableRect> ());
    boost::shared_ptr <DragAndDropInterface> dnd (boost::make_shared <MockDragAndDrop> ());
    boost::shared_ptr <ResizeAllocationInterface> ra (boost::make_shared <MockResizeAllocator> ());

    tl.dragGrabbedWindow (&mr, &mta, &mwm, &md, &mmx, ar, dnd, ra, center);
}

class TestTileLayerDragZones :
    public ::testing::TestWithParam < std::tr1::tuple <CompPoint,
						       CompRect,
						       CompRect> >
{
    protected:

	virtual void SetUp ()
	{
	    std::tr1::tie (toPoint, windowGeometry, dropGeometry) = GetParam ();
	}

    public:

	CompPoint toPoint;
	CompRect windowGeometry;
	CompRect dropGeometry;
};

namespace TestDragToZonesTestData
{
    std::vector <CompPoint> points =
    {
	CompPoint (0, 0),
	CompPoint (0, 999),
	CompPoint (999, 979),
	CompPoint (0, 21)
    };

    std::vector <CompRect> dropGeometries =
    {
	CompRect (0, 0, 100, 100),
	CompRect (10, 10, 200, 200)
    };

    std::vector <CompRect> windowGeometries =
    {
	CompRect (0, 0, 100, 100)
    };
}

ACTION_P (Capture, ptr)
{
    *ptr = arg0;
    return true;
}

TEST_P(TestTileLayerDragZones, DragInDragOut)
{
    MockWindowMonitor mwm;
    MockGrabTracker   mgt;
    StrictMock <MockAssignable> mta;
    MockAnimatableWindow maw;
    MockDrawable md;
    MockDropCommit mdc;
    MockResizable mr;
    MockMaximziationStateApplier mmx;
    FakeTileAssignable fta;
    CompPoint     dragPoint;

    ON_CALL (mta, assignTile (_)).WillByDefault (Invoke (&fta, &FakeTileAssignable::assignTile));
    ON_CALL (mta, releaseTile ()).WillByDefault (Invoke (&fta, &FakeTileAssignable::releaseTile));
    ON_CALL (mta, tile ()).WillByDefault (Invoke (&fta, &FakeTileAssignable::tile));

    TileLayer::GetWorkAreaFunc gwaf (boost::bind (&WindowMonitorInterface::monitorWorkArea, &mwm));

    CompRegion outputr (0, 0, 1000, 1000);
    CompRect workArea (0, 0, 1000, 1000);

    EXPECT_CALL (mwm, monitorRegion ()).WillOnce (Return (outputr));
    EXPECT_CALL (mwm, monitorWorkArea ()).WillRepeatedly (Return (workArea));
    EXPECT_CALL (mgt, addGrabbedWindow (&mta, &maw, &mwm, &md, &mdc, _, dragPoint));

    TileLayer tl (mwm.monitorRegion ().boundingRect (),
		  gwaf,
		  &mgt);

    boost::shared_ptr <MockTile> mnilTile (boost::make_shared <MockTile> ());
    boost::shared_ptr <TileInterface> nilTile (boost::shared_static_cast <TileInterface> (mnilTile));

    EXPECT_CALL (mta, tile ()).WillRepeatedly (ReturnRef (fta.tile ()));

    tl.addGrabbedWindow (&mta, &maw, &mwm, &md, &mdc, gwaf, dragPoint);

    CompRegion noDragZoneReg = outputr;

    for (const CompPoint &p : TestDragToZonesTestData::points)
	noDragZoneReg - tl.dragZoneForPoint (p);

    CompPoint centerRect (noDragZoneReg.boundingRect ().center ());

    CompPoint p (toPoint);

    CompRect r (tl.dragZoneForPoint (p));
    CompPoint center (r.center ());

    EXPECT_CALL (mgt, dragGrabbedWindow (_, _, _, _, _, _, _, _, _));

    boost::shared_ptr <MockAnimatableRect> ar (boost::make_shared <MockAnimatableRect> ());
    boost::shared_ptr <MockDragAndDrop> dnd (boost::make_shared <MockDragAndDrop> ());
    boost::shared_ptr <MockResizeAllocator> ra (boost::make_shared <MockResizeAllocator> ());

    MockDragAndDrop &mdnd = *(dnd.get ());
    MockResizeAllocator &mra = *(ra.get ());
    MockAnimatableRect &mar = *(ar.get ());

    CompRect wGeometry (windowGeometry);

    EXPECT_CALL (mdnd, setGetSnapGeometryFunc (_));
    EXPECT_CALL (mra, tileableFitsAllocation (_)).WillRepeatedly (Return (true));
    EXPECT_CALL (mr, windowGeometry ()).WillRepeatedly (Return (wGeometry));
    EXPECT_CALL (md, getDrawWindowGeometry ()).WillRepeatedly (Return (wGeometry));

    CompRect allocation;

    /* Capture assigned tile here while we don't have a mock for it */
    EXPECT_CALL (mta, assignTile (_));
    EXPECT_CALL (mra, setAllocation (_)).WillOnce (WithArgs<0> (Capture (&allocation)));
    EXPECT_CALL (mra, allocation ()).WillRepeatedly (ReturnRef (allocation));
    EXPECT_CALL (mar, setFromRectFunc (_)).Times (2);
    EXPECT_CALL (mar, setToRectFunc (_)).Times (2);
    EXPECT_CALL (md, damage ()).Times (4);
    EXPECT_CALL (mar, setDirection (_)).Times (2);

    tl.dragGrabbedWindow (&mr, &mta, &mwm, &md, &mmx, ar, dnd, ra, center);

    CompRect dG (dropGeometry);

    /* Snap out window from drag zone */

    EXPECT_CALL (mgt, dragGrabbedWindow (_, _, _, _, _, _, _, _, _));
    EXPECT_CALL (mdnd, setPotentialDropZone (_));
    EXPECT_CALL (mar, progressAnimation (_, _)).Times (2);
    EXPECT_CALL (mdnd, applyDrag (_));
    EXPECT_CALL (md, glDrawSetEnabled (_));
    EXPECT_CALL (mdnd, dropGeometry ()).WillRepeatedly (ReturnRef (dG));
    EXPECT_CALL (mdnd, shouldSnapOut ()).WillOnce (Return (true));

    EXPECT_CALL (mr, applyResizeRules (false));
    EXPECT_CALL (mr, resizeWindow (_));

    tl.dragGrabbedWindow (&mr, &mta, &mwm, &md, &mmx, ar, dnd, ra, centerRect);
}

INSTANTIATE_TEST_CASE_P(TestDragToZones,
			TestTileLayerDragZones,
			Combine (ValuesIn (TestDragToZonesTestData::points),
				 ValuesIn (TestDragToZonesTestData::windowGeometries),
				 ValuesIn (TestDragToZonesTestData::dropGeometries)));

TEST(UnityTileLayer, AddTracks)
{
    MockWindowMonitor mwm;
    MockGrabTracker   mgt;
    MockAssignable mta;
    MockAnimatableWindow maw;
    MockDrawable md;
    MockDropCommit mdc;
    MockResizable mr;
    MockMaximziationStateApplier mmx;
    CompPoint     dragPoint;

    TileLayer::GetWorkAreaFunc gwaf (boost::bind (&WindowMonitorInterface::monitorWorkArea, &mwm));

    UnityTileLayer utl (&mgt);

    EXPECT_CALL (mgt, addGrabbedWindow (&mta,
					&maw,
					&mwm,
					&md,
					&mdc,
					_,
					dragPoint));

    utl.addGrabbedWindow(&mta,
			 &maw,
			 &mwm,
			 &md,
			 &mdc,
			 gwaf,
			 dragPoint);
}

class UnityTileLayerDragTest :
    public ::testing::Test
{
    public:

	UnityTileLayerDragTest () :
	    ar (boost::make_shared <MockAnimatableRect> ()),
	    dnd (boost::make_shared <MockDragAndDrop> ()),
	    ra (boost::make_shared <MockResizeAllocator> ()),
	    gwaf (boost::bind (&WindowMonitorInterface::monitorWorkArea, &mwm)),
	    utl (&mgt),
	    dropMargin (30),
	    workArea (10, 10, 990, 990),
	    noDropArea (workArea.x () + dropMargin,
			workArea.y () + dropMargin,
			workArea.width () - dropMargin * 2,
			workArea.height () - dropMargin),
	    dragPoint (noDropArea.center ())
	{
	    ON_CALL (mta, assignTile (_)).WillByDefault (Invoke (&fta, &FakeTileAssignable::assignTile));
	    ON_CALL (mta, releaseTile ()).WillByDefault (Invoke (&fta, &FakeTileAssignable::releaseTile));
	    ON_CALL (mta, tile ()).WillByDefault (Invoke (&fta, &FakeTileAssignable::tile));
	}

    protected:

	MockWindowMonitor mwm;
	MockGrabTracker   mgt;
	MockAssignable mta;
	MockAnimatableWindow maw;
	MockDrawable md;
	MockDropCommit mdc;
	MockResizable mr;
	MockMaximziationStateApplier mmx;
	boost::shared_ptr <MockAnimatableRect> ar;
	boost::shared_ptr <MockDragAndDrop> dnd;
	boost::shared_ptr <MockResizeAllocator> ra;

	TileLayer::GetWorkAreaFunc gwaf;

	UnityTileLayer utl;

	const unsigned int dropMargin;

	CompRect workArea;
	CompRect noDropArea;

	CompPoint     dragPoint;

	TileInterface::Ptr noTile;

	FakeTileAssignable fta;
};

TEST_F(UnityTileLayerDragTest, DragWithinNotADropZone)
{
    EXPECT_CALL (mwm, monitorWorkArea ()).WillOnce (Return (workArea));
    EXPECT_CALL (mta, tile ()).WillOnce (ReturnRef (noTile));
    EXPECT_CALL (mta, assignTile (_)).Times (0);
    EXPECT_CALL (mgt, dragGrabbedWindow (&mr,
					 &mta,
					 &mwm,
					 &md,
					 &mmx,
					 _,
					 _,
					 _,
					 dragPoint));

    utl.dragGrabbedWindow (&mr,
			   &mta,
			   &mwm,
			   &md,
			   &mmx,
			   ar,
			   dnd,
			   ra,
			   dragPoint);
}

/* FIXME: While we don't have a MockTile, we have
 * to do some needless setup here ... */
TEST_F(UnityTileLayerDragTest, DragOut)
{
    const CompRect topZone (workArea.x (),
			    workArea.y (),
			    workArea.width (),
			    dropMargin);
    const CompRect max (workArea);

    compiz::window::Geometry wg (100, 100, 100, 100, 0);

    dragPoint = topZone.center ();

    EXPECT_CALL (mgt, dragGrabbedWindow (&mr,
					 &mta,
					 &mwm,
					 &md,
					 &mmx,
					 _,
					 _,
					 _,
					 dragPoint));

    EXPECT_CALL (mwm, monitorWorkArea ()).WillRepeatedly (Return (workArea));
    EXPECT_CALL (*ra, tileableFitsAllocation (&mr)).WillOnce (Return (true));
    EXPECT_CALL (mr, windowGeometry ()).WillRepeatedly (Return (wg));
    EXPECT_CALL (md, getDrawWindowGeometry ()).WillRepeatedly (Return (wg));
    EXPECT_CALL (*ra, allocation ()).WillRepeatedly (ReturnRef (max));
    EXPECT_CALL (mta, assignTile (_));

    EXPECT_CALL (*ra, setAllocation (max));
    EXPECT_CALL (*ar, setFromRectFunc (_)).Times (2);
    EXPECT_CALL (*ar, setToRectFunc (_)).Times (2);
    EXPECT_CALL (md, damage ()).Times (6);
    EXPECT_CALL (*ar, setDirection (AnimatableRectInterface::Type::In));
    EXPECT_CALL (*dnd, setGetSnapGeometryFunc (_));
    EXPECT_CALL (mta, tile ()).Times (AtLeast (1));
    EXPECT_CALL (*dnd, setPotentialDropZone (_));

    EXPECT_CALL (*dnd, dropGeometry ()).WillRepeatedly (ReturnRef (wg));

    EXPECT_CALL (*ar, progressAnimation (_, _)).Times (4);
    EXPECT_CALL (*dnd, applyDrag (dragPoint));
    EXPECT_CALL (md, glDrawSetEnabled (_)).Times (2);

    EXPECT_CALL (*dnd, shouldSnapOut ()).WillOnce (Return (false));

    utl.dragGrabbedWindow (&mr,
			   &mta,
			   &mwm,
			   &md,
			   &mmx,
			   ar,
			   dnd,
			   ra,
			   dragPoint);

    CompPoint oldDragPoint (dragPoint);
    dragPoint = noDropArea.center ();

    EXPECT_CALL (mgt, dragGrabbedWindow (_, _, _, _, _, _, _, _, _));
    EXPECT_CALL (*dnd, setPotentialDropZone (_)).Times (0);
    EXPECT_CALL (*dnd, shouldSnapOut ()).WillOnce (Return (true));
    EXPECT_CALL (*dnd, applyDrag (dragPoint - oldDragPoint));
    EXPECT_CALL (*ar, setDirection (AnimatableRectInterface::Type::Out));

    utl.dragGrabbedWindow (&mr,
			   &mta,
			   &mwm,
			   &md,
			   &mmx,
			   ar,
			   dnd,
			   ra,
			   dragPoint);

    /* Tile Destroyed, resize back to normal */
    EXPECT_CALL (mr, applyResizeRules (false));
    EXPECT_CALL (mr, resizeWindow (wg));
}

TEST_F(UnityTileLayerDragTest, DragWithinTopZone)
{
    const CompRect topZone (workArea.x (),
			    workArea.y (),
			    workArea.width (),
			    dropMargin);
    const CompRect max (workArea);

    compiz::window::Geometry wg (100, 100, 100, 100, 0);

    dragPoint = topZone.center ();

    EXPECT_CALL (mgt, dragGrabbedWindow (&mr,
					 &mta,
					 &mwm,
					 &md,
					 &mmx,
					 _,
					 _,
					 _,
					 dragPoint));

    EXPECT_CALL (mwm, monitorWorkArea ()).WillOnce (Return (workArea));
    EXPECT_CALL (mta, tile ()).Times (AtLeast (1));
    EXPECT_CALL (*ra, tileableFitsAllocation (&mr)).WillOnce (Return (true));
    EXPECT_CALL (mr, windowGeometry ()).WillRepeatedly (Return (wg));
    EXPECT_CALL (*ra, allocation ()).WillRepeatedly (ReturnRef (max));
    EXPECT_CALL (mta, assignTile (_));

    EXPECT_CALL (*ra, setAllocation (max));
    EXPECT_CALL (*ar, setFromRectFunc (_));
    EXPECT_CALL (*ar, setToRectFunc (_));
    EXPECT_CALL (md, damage ()).Times (3);
    EXPECT_CALL (*ar, setDirection (AnimatableRectInterface::Type::In));
    EXPECT_CALL (*dnd, setGetSnapGeometryFunc (_));
    EXPECT_CALL (*dnd, setPotentialDropZone (_));
    EXPECT_CALL (*ar, progressAnimation (_, _)).Times (2);
    EXPECT_CALL (*dnd, dropGeometry ()).WillRepeatedly (ReturnRef (wg));
    EXPECT_CALL (*dnd, applyDrag (dragPoint));
    EXPECT_CALL (md, glDrawSetEnabled (_)).Times (1);
    EXPECT_CALL (*dnd, shouldSnapOut ());

    utl.dragGrabbedWindow (&mr,
			   &mta,
			   &mwm,
			   &md,
			   &mmx,
			   ar,
			   dnd,
			   ra,
			   dragPoint);

    /* Tile Destroyed, resize back to normal */
    EXPECT_CALL (mr, applyResizeRules (false));
    EXPECT_CALL (mr, resizeWindow (wg));
}

TEST_F(UnityTileLayerDragTest, DragWithinLeftZone)
{
    const CompRect leftZone (workArea.x (),
			     workArea.y (),
			     dropMargin,
			     workArea.height ());
    const CompRect lMax (workArea.x (),
			 workArea.y (),
			 workArea.width () / 2,
			 workArea.height ());

    compiz::window::Geometry wg (100, 100, 100, 100, 0);

    dragPoint = leftZone.center ();

    EXPECT_CALL (mgt, dragGrabbedWindow (&mr,
					 &mta,
					 &mwm,
					 &md,
					 &mmx,
					 _,
					 _,
					 _,
					 dragPoint));

    EXPECT_CALL (mwm, monitorWorkArea ()).WillOnce (Return (workArea));
    EXPECT_CALL (mta, tile ()).Times (AtLeast (1));
    EXPECT_CALL (*ra, tileableFitsAllocation (&mr)).WillOnce (Return (true));
    EXPECT_CALL (mr, windowGeometry ()).WillRepeatedly (Return (wg));
    EXPECT_CALL (*ra, allocation ()).WillRepeatedly (ReturnRef (lMax));
    EXPECT_CALL (mta, assignTile (_));

    EXPECT_CALL (*ra, setAllocation (lMax));
    EXPECT_CALL (*ar, setFromRectFunc (_));
    EXPECT_CALL (*ar, setToRectFunc (_));
    EXPECT_CALL (md, damage ()).Times (3);
    EXPECT_CALL (*ar, setDirection (AnimatableRectInterface::Type::In));
    EXPECT_CALL (*dnd, setGetSnapGeometryFunc (_));
    EXPECT_CALL (*dnd, setPotentialDropZone (_));
    EXPECT_CALL (*ar, progressAnimation (_, _)).Times (2);
    EXPECT_CALL (*dnd, dropGeometry ()).WillRepeatedly (ReturnRef (wg));
    EXPECT_CALL (*dnd, applyDrag (dragPoint));
    EXPECT_CALL (md, glDrawSetEnabled (_)).Times (1);
    EXPECT_CALL (*dnd, shouldSnapOut ());

    utl.dragGrabbedWindow (&mr,
			   &mta,
			   &mwm,
			   &md,
			   &mmx,
			   ar,
			   dnd,
			   ra,
			   dragPoint);

    /* Tile Destroyed, resize back to normal */
    EXPECT_CALL (mr, applyResizeRules (false));
    EXPECT_CALL (mr, resizeWindow (wg));
}

TEST_F(UnityTileLayerDragTest, DragWithinRightZone)
{
    const CompRect rightZone (workArea.x2 () - dropMargin,
			      workArea.y (),
			      dropMargin,
			      workArea.height ());
    const CompRect rMax (workArea.x () + workArea.width () / 2,
			 workArea.y (),
			 workArea.width () / 2,
			 workArea.height ());

    compiz::window::Geometry wg (100, 100, 100, 100, 0);

    dragPoint = rightZone.center ();

    EXPECT_CALL (mgt, dragGrabbedWindow (&mr,
					 &mta,
					 &mwm,
					 &md,
					 &mmx,
					 _,
					 _,
					 _,
					 dragPoint));

    EXPECT_CALL (mwm, monitorWorkArea ()).WillOnce (Return (workArea));
    EXPECT_CALL (mta, tile ()).Times (AtLeast (1));
    EXPECT_CALL (*ra, tileableFitsAllocation (&mr)).WillOnce (Return (true));
    EXPECT_CALL (mr, windowGeometry ()).WillRepeatedly (Return (wg));
    EXPECT_CALL (*ra, allocation ()).WillRepeatedly (ReturnRef (rMax));

    EXPECT_CALL (mta, assignTile (_));

    EXPECT_CALL (*ra, setAllocation (rMax));
    EXPECT_CALL (*ar, setFromRectFunc (_));
    EXPECT_CALL (*ar, setToRectFunc (_));
    EXPECT_CALL (md, damage ()).Times (3);
    EXPECT_CALL (*ar, setDirection (AnimatableRectInterface::Type::In));
    EXPECT_CALL (*dnd, setGetSnapGeometryFunc (_));
    EXPECT_CALL (*dnd, setPotentialDropZone (_));
    EXPECT_CALL (*ar, progressAnimation (_, _)).Times (2);
    EXPECT_CALL (*dnd, dropGeometry ()).WillRepeatedly (ReturnRef (wg));
    EXPECT_CALL (*dnd, applyDrag (dragPoint));
    EXPECT_CALL (md, glDrawSetEnabled (_)).Times (1);
    EXPECT_CALL (*dnd, shouldSnapOut ());

    utl.dragGrabbedWindow (&mr,
			   &mta,
			   &mwm,
			   &md,
			   &mmx,
			   ar,
			   dnd,
			   ra,
			   dragPoint);

    /* Tile Destroyed, resize back to normal */
    EXPECT_CALL (mr, applyResizeRules (false));
    EXPECT_CALL (mr, resizeWindow (wg));
}

TEST(UnityTileLayer, RemoveUntracks)
{
    MockGrabTracker   mgt;
    MockDropCommit mdc;

    UnityTileLayer utl (&mgt);

    EXPECT_CALL (mgt, removeGrabbedWindow (&mdc));

    utl.removeGrabbedWindow (&mdc);
}

TEST_F(UnityTileLayerDragTest, DragThenDropResizesResizable)
{
    /* Drag it to the top zone */
    const CompRect topZone (workArea.x (),
			    workArea.y (),
			    workArea.width (),
			    dropMargin);
    const CompRect max (workArea);

    compiz::window::Geometry wg (100, 100, 100, 100, 0);

    dragPoint = topZone.center ();

    EXPECT_CALL (mgt, dragGrabbedWindow (&mr,
					 &mta,
					 &mwm,
					 &md,
					 &mmx,
					 _,
					 _,
					 _,
					 dragPoint));

    EXPECT_CALL (mwm, monitorWorkArea ()).WillOnce (Return (workArea));
    EXPECT_CALL (mta, tile ()).Times (AtLeast (1));
    EXPECT_CALL (*ra, tileableFitsAllocation (&mr)).WillOnce (Return (true));
    EXPECT_CALL (mr, windowGeometry ()).WillRepeatedly (Return (wg));
    EXPECT_CALL (*ra, allocation ()).WillRepeatedly (ReturnRef (max));
    EXPECT_CALL (mta, assignTile (_));

    EXPECT_CALL (*ra, setAllocation (max));
    EXPECT_CALL (*ar, setFromRectFunc (_));
    EXPECT_CALL (*ar, setToRectFunc (_));
    EXPECT_CALL (md, damage ()).Times (3);
    EXPECT_CALL (*ar, setDirection (AnimatableRectInterface::Type::In));
    EXPECT_CALL (*dnd, setGetSnapGeometryFunc (_));
    EXPECT_CALL (*dnd, setPotentialDropZone (_));
    EXPECT_CALL (*ar, progressAnimation (_, _)).Times (2);
    EXPECT_CALL (*dnd, dropGeometry ()).WillRepeatedly (ReturnRef (wg));
    EXPECT_CALL (*dnd, applyDrag (dragPoint));
    EXPECT_CALL (md, glDrawSetEnabled (_)).Times (1);
    EXPECT_CALL (*dnd, shouldSnapOut ());

    utl.dragGrabbedWindow (&mr,
			   &mta,
			   &mwm,
			   &md,
			   &mmx,
			   ar,
			   dnd,
			   ra,
			   dragPoint);

    /* All removeGrabbedWindow does is remove the grab
     * it is up to the grab tracker implementation to do
     * the rest */
    EXPECT_CALL (mgt, removeGrabbedWindow (&mdc));

    utl.removeGrabbedWindow (&mdc);

    /* This test is currently a little awkward because we need to
     * fetch the constructed tile from our fake assignable */
    const TileInterface::Ptr &tile (fta.tile ());

    /* This should forward on to lockTargetSize */
    EXPECT_CALL (*dnd, ungrab ());
    EXPECT_CALL (*ar, setDirection (AnimatableRectInterface::Type::Complete));
    EXPECT_CALL (*ar, progressAnimation (0, true));
    /* The animation is complete */
    EXPECT_CALL (*ar, tweenedGeometry (wg)).WillOnce (Return (max));
    EXPECT_CALL (mr, toWindowGeometry (max)).WillOnce (Return (max));
    EXPECT_CALL (mr, applyResizeRules (true));
    EXPECT_CALL (mmx, applyMaximizeState (max));
    EXPECT_CALL (mr, resizeWindow (max));

    tile->dropCommit ();

    /* Tile Destroyed, resize back to normal */
    EXPECT_CALL (mr, applyResizeRules (false));
    EXPECT_CALL (mr, resizeWindow (wg));
}
