/*
 * Compiz Tiles Plugin
 *
 * tiles.cpp
 *
 * Copyright (c) 2011 Sam Spilsbury <smspillaz@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include "tiles.h"
#include <boost/assign/std/vector.hpp>
#include <boost/math/distributions/normal.hpp>
#include <boost/make_shared.hpp>

using namespace boost::assign;
using boost::math::normal;

COMPIZ_PLUGIN_20090315 (tiles, TilesPluginVTable);

class PrivateTilesWindow
{
    public:

	PrivateTilesWindow (FramebufferUpdateQueueInterface *,
			    FramebufferUpdateInterface *,
			    WindowMonitorInterface *wm,
			    ResizableInterface *rs,
			    MaximizableInterface *mi);

	CompWindow      *window;
	CompositeWindow *cWindow;
	GLWindow        *gWindow;

	CompRect        mGrabGeometry;

	TileInterface::Ptr       mTile;

	unsigned int    mGrabMask;
	CompRect        mResizeRectangle;

	BlurBackingFramebuffer::Ptr mVPassBuffer;
	BlurBackingFramebuffer::Ptr mHPassBuffer;

	GrabTrackingInterface *mGrabTracker;
	GrabTrackerFinderInterface *mGrabTrackerFinder;
	PointerFetchInterface *mPointerFetch;

	UpdatableFramebufferForwarder mFramebufferForwarder;
	UpdatableFramebufferCommunicator mFramebufferCommunicator;

	unsigned int         mSavedSizeHintsFlags;

	MaximizationStateApplier mMaximizer;

	bool mGlPaintEnabled;
};

struct DraggedWindowInfo
{
    TileAssignableInterface *assignable;
    AnimatableWindowInterface *animatable;
    WindowMonitorInterface *monitor;
    DrawableWindowInterface *drawable;
    DropCommitInterface *dropCommit;
};

class PrivateTilesScreen
{
    public:

	CompositeScreen *cScreen;
	GLScreen        *gScreen;

	std::vector <DraggedWindowInfo> mDraggedWindows;
	std::vector <DropCommitInterface *> mResizedWindows;
	std::vector <TileLayerTrackingGrabsInterface::Ptr> mTileLayers;

	constexpr static
	const char * blurVertexProgram =
	    "varying vec2 vTexCoord0;            \n"\
	    "attribute vec2 texCoord0;           \n"\
	    "attribute vec3 position;		 \n"\
	    "uniform mat4 projection;            \n"\
	    "uniform mat4 modelview;             \n"\
	    "void main()                         \n"\
	    "{                                   \n"\
	    "  vTexCoord0 = texCoord0;           \n"\
	    "  gl_Position = projection * modelview * vec4 (position, 1.0);\n"\
	    "}";


	constexpr static
	const char * hPassFragmentProgram =
	    "uniform sampler2D texture0;                             \n"\
	    "uniform vec2 TextureSize0;                                    \n"\
	    "varying vec2 vTexCoord0; \n"\
	    "vec4 SampleTexture(sampler2D TexObject, vec2 TexCoord)       \n"\
	    "{                                                             \n"\
	    " return texture2D(TexObject, TexCoord.st);                  \n"\
	    "}                                                             \n"\
	    "#define NUM_SAMPLES 7                                         \n"\
	    "uniform float W0;\n"\
	    "uniform float W1;\n"\
	    "uniform float W2;\n"\
	    "uniform float W3;\n"\
	    "uniform float W4;\n"\
	    "uniform float W5;\n"\
	    "uniform float W6;\n"\
	    "void main()                                                  \n"\
	    "{                                                             \n"\
	    "  vec4 sum   = vec4(0.0, 0.0, 0.0, 0.0);                     \n"\
	    "  vec2 delta = vec2(1.0 / TextureSize0.x, 0.0);              \n"\
	    "  vec2 texCoord = vec2(vTexCoord0.s, vTexCoord0.t);          \n"\
	    "  texCoord.x -= float((NUM_SAMPLES - 1) / 2) / TextureSize0.x; \n"\
	    "  sum += SampleTexture(texture0, texCoord) * W0;     \n"\
	    "  texCoord += delta;                                          \n"\
	    "  sum += SampleTexture(texture0, texCoord) * W1;     \n"\
	    "  texCoord += delta;                                          \n"\
	    "  sum += SampleTexture(texture0, texCoord) * W2;     \n"\
	    "  texCoord += delta;                                          \n"\
	    "  sum += SampleTexture(texture0, texCoord) * W3;     \n"\
	    "  texCoord += delta;                                          \n"\
	    "  sum += SampleTexture(texture0, texCoord) * W4;     \n"\
	    "  texCoord += delta;                                          \n"\
	    "  sum += SampleTexture(texture0, texCoord) * W5;     \n"\
	    "  texCoord += delta;                                          \n"\
	    "  sum += SampleTexture(texture0, texCoord) * W6;     \n"\
	    "  texCoord += delta;                                          \n"\
	    "  gl_FragColor = vec4(sum.x, sum.y, sum.z, sum.w);           \n"\
	    "}";

	constexpr static
	const char * vPassFragmentProgram =
	    "uniform sampler2D texture0;                             \n"\
	    "uniform vec2 TextureSize0;                                    \n"\
	    "varying vec2 vTexCoord0; \n"\
	    "vec4 SampleTexture(sampler2D TexObject, vec2 TexCoord)       \n"\
	    "{                                                             \n"\
	    " return texture2D(TexObject, TexCoord.st);                  \n"\
	    "}                                                             \n"\
	    "#define NUM_SAMPLES 7                                         \n"\
	    "uniform float W0;\n"\
	    "uniform float W1;\n"\
	    "uniform float W2;\n"\
	    "uniform float W3;\n"\
	    "uniform float W4;\n"\
	    "uniform float W5;\n"\
	    "uniform float W6;\n"\
	    "void main()                                                  \n"\
	    "{                                                             \n"\
	    "  vec4 sum   = vec4(0.0, 0.0, 0.0, 0.0);                     \n"\
	    "  vec2 delta = vec2(0.0, 1.0 / TextureSize0.y);              \n"\
	    "  vec2 texCoord = vec2(vTexCoord0.s, vTexCoord0.t);          \n"\
	    "  texCoord.y -= float((NUM_SAMPLES - 1) / 2) / TextureSize0.y;     \n"\
	    "  sum += SampleTexture(texture0, texCoord) * W0;     \n"\
	    "  texCoord += delta;                                          \n"\
	    "  sum += SampleTexture(texture0, texCoord) * W1;     \n"\
	    "  texCoord += delta;                                          \n"\
	    "  sum += SampleTexture(texture0, texCoord) * W2;     \n"\
	    "  texCoord += delta;                                          \n"\
	    "  sum += SampleTexture(texture0, texCoord) * W3;     \n"\
	    "  texCoord += delta;                                          \n"\
	    "  sum += SampleTexture(texture0, texCoord) * W4;     \n"\
	    "  texCoord += delta;                                          \n"\
	    "  sum += SampleTexture(texture0, texCoord) * W5;     \n"\
	    "  texCoord += delta;                                          \n"\
	    "  sum += SampleTexture(texture0, texCoord) * W6;     \n"\
	    "  texCoord += delta;                                          \n"\
	    "  gl_FragColor = vec4(sum.x, sum.y, sum.z, sum.w);           \n"\
	    "}";

	GLProgram *mVPassProgram;
	GLProgram *mHPassProgram;

	GLint mVPassFragmentShader;
	GLint mHPassFragmentShader;
	GLint mBlurVertexShader;

	CompPoint mPointerPosition;

	FramebufferUpdateQueue mFramebufferUpdateQueue;

	CompOutput *mCurrentlyPaintedOutput;
};


static bool veceq (const GLVector &lhs, const GLVector &rhs)
{
    return lhs[GLVector::x] == rhs[GLVector::x] &&
	    lhs[GLVector::y] == rhs[GLVector::y] &&
	    lhs[GLVector::z] == rhs[GLVector::z] &&
	    lhs[GLVector::w] == rhs[GLVector::w];
}






bool
TilesWindow::assignTile (TileInterface::Ptr tp)
{
    priv->mTile = tp;
    return true;
}



void
TilesWindow::releaseTile ()
{
    /* Remove all framebuffer updates */
    priv->mFramebufferCommunicator.dequeueUpdates ();

    glDrawSetEnabled (false);
    priv->mTile.reset ();
}

void
TilesWindow::dropCommit ()
{
    if (priv->mTile)
	priv->mTile->dropCommit ();

    /* Remove all frameubffer updates */
    priv->mFramebufferCommunicator.dequeueUpdates ();

    /* For now */
    glDrawSetEnabled (false);
}

void
TilesWindow::setGrabTracker(GrabTrackingInterface *grabTracker)
{
    priv->mGrabTracker = grabTracker;
}

const TileInterface::Ptr &
TilesWindow::tile () const
{
    return priv->mTile;
}

CompRect
TilesWindow::grabGeometry ()
{
    return priv->mGrabGeometry;
}

CompRect
TilesWindow::getDrawWindowGeometry ()
{
    if (priv->mGrabMask & (CompWindowGrabResizeMask))
    {
	return priv->mResizeRectangle;
    }
    else
	return priv->window->serverBorderRect ();
}

CompRect
TilesWindow::frameGeometry ()
{
    return priv->window->serverBorderRect ();
}

CompRect
TilesWindow::windowGeometry ()
{
    return priv->window->serverGeometry ();
}

void
TilesWindow::resizeWindow (const CompRect &resizeTarget)
{
    XWindowChanges xwc;
    unsigned int   valueMask;

    valueMask = CWX | CWY | CWWidth | CWHeight;
    xwc.x = resizeTarget.x ();
    xwc.y = resizeTarget.y ();
    xwc.width = resizeTarget.width ();
    xwc.height = resizeTarget.height ();

    priv->window->configureXWindow (valueMask, &xwc);
}

CompRect
TilesWindow::toWindowGeometry (const CompRect &frameTarget)
{
    CompRect r (frameTarget);

    r.setX (r.x () + priv->window->border ().left);
    r.setWidth (r.width () - priv->window->border ().right
		- priv->window->border ().left);
    r.setY (r.y () + priv->window->border ().top);
    r.setHeight (r.height () - priv->window->border ().bottom
		 - priv->window->border ().top);

    return r;
}

CompRect
TilesWindow::toFrameGeometry (const CompRect &windowTarget)
{
    CompRect r (windowTarget);

    r.setX (r.x () - priv->window->border ().left);
    r.setWidth (r.width () + priv->window->border ().right
		+ priv->window->border ().left);
    r.setY (r.y () - priv->window->border ().top);
    r.setHeight (r.height () + priv->window->border ().bottom
		 + priv->window->border ().top);

    return r;
}

void
TilesWindow::applyMaximizeState (const CompRect &resizeTarget)
{
    priv->mMaximizer.applyMaximizeState (resizeTarget);
}

void
TilesWindow::maximize (unsigned int s)
{
    priv->window->maximize (s);
}

void
TilesWindow::applyResizeRules (bool lock)
{
    if (lock)
    {
	if (priv->window->sizeHints ().flags & PResizeInc)
	{
	    priv->mSavedSizeHintsFlags |= PResizeInc;
	    priv->window->sizeHints ().flags &= ~(PResizeInc);
	}
	else
	    priv->window->sizeHints ().flags = 0;
    }
    else
    {
	priv->window->sizeHints ().flags |= priv->mSavedSizeHintsFlags;
	priv->mSavedSizeHintsFlags = 0;
    }
}

void
TilesWindow::glDrawSetEnabled (bool k)
{
    priv->gWindow->glPaintSetEnabled (this, k);
    priv->mGlPaintEnabled = k;
}

CompRect
TilesWindow::monitorWorkArea ()
{
    return screen->outputDevs ()[priv->window->outputDevice ()].workArea ();
}

CompRegion
TilesWindow::monitorRegion ()
{
    return CompRegion (screen->outputDevs ()[priv->window->outputDevice ()]);
}

void
TilesWindow::grabNotify (int x, int y, unsigned int state, unsigned int mask)
{
    priv->window->grabNotify (x, y, state, mask);

    priv->mGrabMask = mask;
    priv->mGrabGeometry = priv->window->serverBorderRect ();

    if (mask & (CompWindowGrabMoveMask))
    {
	TileInterface::Ptr tp (priv->mTile);

	priv->mGrabTracker = priv->mGrabTrackerFinder->findTracker (this,
								    priv->mPointerFetch->getPointerPosition ());
	if (!priv->mGrabTracker->addGrabbedWindow (this,
						   this,
						   this,
						   this,
						   this,
						   boost::bind (&WindowMonitorInterface::monitorWorkArea, this),
						   CompPoint (pointerX, pointerY)))
	{
	    priv->mGrabTracker = NULL;
	    return;
	}
	else
	{
	    if (!tp)
		priv->window->moveNotifySetEnabled (this, true);
	}
    }
    else if (mask & (CompWindowGrabResizeMask) &&
	     priv->mTile)
    {
	priv->mResizeRectangle = priv->window->serverBorderRect ();
	TilesScreen::get (screen)->addResizingWindow (this);
	priv->gWindow->glPaintSetEnabled (this, false);
    }
}

void
TilesWindow::ungrabNotify ()
{
    if (priv->mGrabMask & CompWindowGrabResizeMask &&
	priv->mTile)
    {
	priv->mTile->setTarget (priv->mResizeRectangle);
	priv->mTile->setAllocation (priv->mResizeRectangle);
    }

    TilesScreen::get (screen)->removeGrabbedWindow (this);

    priv->mGrabMask = 0;
    priv->mGrabTracker = NULL;

    if (!priv->mTile)
	priv->window->moveNotifySetEnabled (this, false);

    priv->window->ungrabNotify ();

    priv->gWindow->glPaintSetEnabled (this, false);
    priv->mFramebufferCommunicator.dequeueUpdates ();
}

void
TilesWindow::stateChangeNotify (unsigned int lastState)
{
    TileLayerMiscInterface::WindowState lostState = 0;

    if (lastState & CompWindowStateMaximizedVertMask &&
	!(priv->window->state () & CompWindowStateMaximizedVertMask))
	lostState |= TileLayerMiscInterface::VerticallyMaximized;

    if (lastState & CompWindowStateMaximizedHorzMask &&
	!(priv->window->state () & CompWindowStateMaximizedHorzMask))
	lostState |= TileLayerMiscInterface::HorizontallyMaximized;

    if (lostState)
	TilesScreen::get (screen)->handleStateLoss (this, lostState);

    /* re-track this window */
    if (!priv->mGrabTracker &&
	priv->mGrabMask & CompWindowGrabMoveMask)
    {
	priv->mGrabTracker =
	    priv->mGrabTrackerFinder->findTracker (this,
						   CompPoint (pointerX,
							      pointerY));

	if (priv->mGrabTracker)
	{
	    if (!priv->mGrabTracker->addGrabbedWindow (this,
						       this,
						       this,
						       this,
						       this,
						       boost::bind (&WindowMonitorInterface::monitorWorkArea, this),
						       CompPoint (pointerX, pointerY)))
		priv->mGrabTracker = NULL;
	    else
	    {
		if (!priv->mTile)
		    priv->window->moveNotifySetEnabled (this, true);
	    }
	}
    }

    priv->window->stateChangeNotify (lastState);
}

void
TilesWindow::moveNotify (int dx, int dy, bool immediate)
{
    if (priv->mGrabMask & CompWindowGrabMoveMask &&
	priv->mGrabTracker)
    {
	boost::shared_ptr <DragAndDrop> dnd (boost::make_shared<DragAndDrop> ());
	boost::shared_ptr <ResizeAllocator> ra (boost::make_shared<ResizeAllocator> (this));
	boost::shared_ptr <LinearAnimatedRect> lar (boost::make_shared <LinearAnimatedRect> ());
    	priv->mGrabTracker->dragGrabbedWindow (this,
					       this,
					       this,
					       this,
					       this,
					       boost::shared_static_cast <AnimatableRectInterface> (lar),
					       dnd,
					       boost::shared_static_cast <ResizeAllocationInterface> (ra),
					       CompPoint (pointerX, pointerY));
    }

    priv->window->moveNotify (dx, dy, immediate);
}

void
TilesWindow::validateResizeRequest (unsigned int &mask,
				    XWindowChanges *xwc,
				    unsigned int source)
{
    if (priv->mTile)
    {
    	CompRect actual = priv->mTile->allocation ();

    	priv->mTile->resized (actual);
	/* force xwc to be the current geometry */
	xwc->x = actual.x ();
	xwc->y = actual.y ();
	xwc->width = actual.width ();
	xwc->height = actual.height ();

	mask |= CWX | CWY | CWWidth | CWHeight;
    }

    priv->window->validateResizeRequest (mask, xwc, source);
}

void
TilesWindow::onResizeNotification (int x, int y,
				   int width, int height)
{
    if (priv->mTile)
    {
	CompRect r (x - priv->window->border ().left,
		    y - priv->window->border ().top,
		    width + priv->window->border ().left + priv->window->border ().right,
		    height + priv->window->border ().top + priv->window->border ().bottom);

	priv->mResizeRectangle = r;

	priv->mTile->setTarget (r);
	priv->mTile->setAllocation (r);
	priv->mTile->setAnimation (AnimatableRectInterface::Type::Complete);
    }
}

bool
TilesWindow::fboUpdate ()
{
    if (!priv->mTile)
    	return false;

    GLWindowPaintAttrib wAttrib;
    GLMatrix            wTransform;
    CompRect outRect = priv->window->serverOutputRect ();
    unsigned int        mask = priv->gWindow->lastMask () & 0x2000;

    wAttrib.opacity = OPAQUE;
    wAttrib.brightness = BRIGHT;
    wAttrib.saturation = COLOR;

    mask |= PAINT_WINDOW_TRANSFORMED_MASK;

    /* Expand by about 5 on either side */
    outRect.setLeft (outRect.left ());
    outRect.setRight (outRect.right ());
    outRect.setTop (outRect.top ());
    outRect.setBottom (outRect.bottom ());

    unsigned int oldGlAddGeometryIndex = priv->gWindow->glAddGeometryGetCurrentIndex ();
    priv->gWindow->glAddGeometrySetCurrentIndex (MAXSHORT);

    CompSize bSize (outRect.width (), outRect.height ());

    /* Create framebuffers if they don't exist ... */
    if (!priv->mVPassBuffer ||
    	!priv->mHPassBuffer ||
    	priv->mVPassBuffer->size () != bSize ||
    	priv->mHPassBuffer->size () != bSize)
    {
    	try
    	{
	    TilesScreen *ts = TilesScreen::get (screen);
	    priv->mVPassBuffer = BlurBackingFramebuffer::create (bSize,
								 ts->vPassProgram (),
								 priv->gWindow);
	    priv->mHPassBuffer = BlurBackingFramebuffer::create (bSize,
								 ts->hPassProgram (),
								 priv->gWindow);
    	}
    	catch (...)
    	{
	    priv->mVPassBuffer.reset ();
	    priv->mHPassBuffer.reset ();
    	}
    }

    try
    {
	/* Copy the output - we need to modify some parameters */
	CompOutput        output = screen->outputDevs() [0];
	GLMatrix          identity;
	CompRegion        geometry (0, 0, bSize.width (), bSize.height ());

	/* The output is always at 0, 0 - but it keeps the same
	 * dimentions */
	output.setPos (CompPoint (0, 0));

	wTransform.toScreenSpace (&output, -DEFAULT_Z_CAMERA);
	identity.toScreenSpace (&output, -DEFAULT_Z_CAMERA);

	identity.translate (0.0f, output.height () - outRect.height (), 0.0f);

	FramebufferBinder vbuffer (priv->mVPassBuffer,
				   output);
	FramebufferPainter hpainter (priv->mHPassBuffer,
				     output,
				     1.0f,
				     identity,
				     geometry.boundingRect (),
				     geometry.boundingRect (),
				     1.0f,
				     geometry);
	FramebufferBinder hbuffer (priv->mHPassBuffer,
				   output);

	wTransform.translate (-(outRect.x ()),
			      -(outRect.y ()) + output.height () - outRect.height (),
			      0.0f);

	priv->gWindow->glDraw (wTransform, wAttrib, infiniteRegion, mask);
    }
    catch (...)
    {
    }

    priv->gWindow->glAddGeometrySetCurrentIndex (oldGlAddGeometryIndex);
    return true;
}

void
TilesWindow::setFramebufferHolder (FramebufferUpdateQueueInterface *updateQueue)
{
    priv->mFramebufferCommunicator.setFramebufferHolder (updateQueue);
}

void
TilesWindow::queueUpdate ()
{
    if (priv->mTile &&
    	priv->mGrabMask & CompWindowGrabMoveMask)
    {
	glDrawSetEnabled (true);
	priv->mFramebufferCommunicator.queueUpdate ();
    }
}

void
TilesWindow::dequeueUpdates ()
{
    return priv->mFramebufferCommunicator.dequeueUpdates  ();
}

bool TilesScreen::addQueuedUpdate(FramebufferUpdateInterface *update)
{
    if (priv->mFramebufferUpdateQueue.addQueuedUpdate (update))
    {
	priv->cScreen->preparePaintSetEnabled (this, true);
	return true;
    }

    return false;
}

void
TilesScreen::removeUpdatesFromQueue (FramebufferUpdateInterface *update)
{
    priv->mFramebufferUpdateQueue.removeUpdatesFromQueue (update);
}

void
TilesScreen::dispatch ()
{
    priv->mFramebufferUpdateQueue.dispatch ();
}

bool
TilesScreen::pending ()
{
    return priv->mFramebufferUpdateQueue.pending ();
}

bool
TilesWindow::glPaint (const GLWindowPaintAttrib &attrib, const GLMatrix &matrix,
		      const CompRegion &region, unsigned int mask)
{
    if (priv->mTile)
    {
	GLMatrix wTransform (matrix);
	const CompRect &target = priv->mTile->target ();
	GLfloat alpha = 1.0f;
	CompRect rect = getDrawWindowGeometry ();
	GLVector scale = priv->mTile->scale (rect);
	
	GLWindowPaintAttrib wTransformedAttrib (attrib);
	GLWindowPaintAttrib wAttrib (attrib);

	if (priv->mGrabMask & CompWindowGrabMoveMask)
	{
	    /* When we moving from free to tiled
	     * (eg, actual geometry != target, show
	     *  both the preview and the window, otherwise
	     *  fade through) */

	    if (target == priv->mTile->allocation ())
	    {
		wAttrib.opacity = OPAQUE - (wAttrib.opacity * 0.15);
		alpha *= ((priv->mTile->opacity () * 0.68) / static_cast <float> (OPAQUE));
	    }
	    else
		wAttrib.opacity = OPAQUE - (wAttrib.opacity);

	    mask |= PAINT_WINDOW_TRANSLUCENT_MASK;
	    mask |= PAINT_WINDOW_BLEND_MASK;
	}

	if (priv->mGrabMask & CompWindowGrabMoveMask)
	{
	    GLMatrix identity1;
	    CompOutput        *output = TilesScreen::get (screen)->currentlyPaintedOutput ();

	    identity1.toScreenSpace (output, -DEFAULT_Z_CAMERA);

	    try
	    {
		std::vector <GLVector> geometry;
		CompRect sbr = priv->window->serverBorderRect ();
		CompRect current = priv->mTile->current (sbr);

		FramebufferPainter vpainter (priv->mVPassBuffer,
					     *output,
					     (priv->mTile->opacity () / static_cast <GLfloat> (OPAQUE)),
					     identity1,
					     current,
					     sbr,
					     alpha,
					     region);
	    }
	    catch (...)
	    {
		mask |= PAINT_WINDOW_TRANSFORMED_MASK;
		GLVector translation (priv->mTile->translation (rect));

		/* The fallback mode doesn't quite work right at the moment, this transformation
		 * should fit inside the bounding box but doesn't */
	    	wTransform.scale (scale);
		wTransform.translate (translation);
	    	wTransformedAttrib.opacity = alpha * OPAQUE;
		priv->gWindow->glPaint (wTransformedAttrib, wTransform, region, mask);
	    }

	    /* XXX: if we make this semitransparent, there's a nasty state leak .... */
	    return priv->gWindow->glPaint (attrib, matrix, region, mask);
	}
	else if (scale[GLVector::x] != 1.0f ||
		 scale[GLVector::y] != 1.0f ||
		 scale[GLVector::z] != 1.0f)
	{
	    wTransform.scale (scale);

	    mask |= PAINT_WINDOW_TRANSFORMED_MASK;

	    return priv->gWindow->glPaint (wAttrib, wTransform, region, mask);
	}
    }

    return priv->gWindow->glPaint (attrib, matrix, region, mask);
}

bool
TilesWindow::damageRect (bool initial, const CompRect &rect)
{
    bool ret = priv->cWindow->damageRect (initial, rect);

    if (priv->mTile)
    {
	CompRect wg (priv->window->serverGeometry ());

	const CompRect &current = priv->mTile->current (priv->window->serverBorderRect ());
	CompRect currentWithoutDecorations (toWindowGeometry (current));
	CompRect r (scaleChild (CompRect (rect.x () + wg.x (),
					  rect.y () + wg.y (),
					  rect.width (),
					  rect.height ()), wg, currentWithoutDecorations)); // target relative

	/* Twenty pixels padding */
	r.setLeft (r.left () - 20);
	r.setRight (r.right () + 20);
	r.setTop (r.top () - 20);
	r.setBottom (r.bottom () + 20);

	CompositeScreen::get (screen)->damageRegion (CompRegion (r));
    }

    /* Allow core to damage the rest of the window */
    return ret;
}

void
TilesWindow::windowNotify(CompWindowNotify n)
{
    switch (n)
    {
      case CompWindowNotifyClose:
	if (priv->mTile &&
	    priv->mTile->layer ())
	    priv->mTile->layer ()->removeWindow (this);
      default:
	break;
    }

    priv->window->windowNotify (n);
}

bool
TilesWindow::animate (unsigned int ms)
{
    if (priv->mTile)
    {
	bool ret = priv->mTile->animate (ms);

	if (ret)
	    queueUpdate ();

	return ret;
    }
    else
	return true;
}

void
TilesWindow::damage ()
{
    CompositeScreen *cScreen = CompositeScreen::get (screen);
    TilesScreen     *tScreen = TilesScreen::get (screen);

    cScreen->preparePaintSetEnabled (tScreen, true);
    cScreen->donePaintSetEnabled (tScreen, true);

    priv->cWindow->addDamage (true);
}

GLProgram *
TilesScreen::hPassProgram () const
{
    return priv->mHPassProgram;
}

GLProgram *
TilesScreen::vPassProgram () const
{
    return priv->mVPassProgram;
}

CompOutput *
TilesScreen::currentlyPaintedOutput () const
{
    return priv->mCurrentlyPaintedOutput;
}

void
TilesScreen::handleWindowResized (TilesWindow *tw, int x, int y, int width, int height)
{
    /* TileLayer::handleWindowResize needs to know
     * the old locked size first in order to resize
     * the other tiles, so process this first */
    if (priv->mTileLayers.size ())
    {
	TileLayerTrackingGrabsInterface::Ptr tp = priv->mTileLayers.front (); // assuming here that there is a tileLayer since there is a tile

	tp->handleWindowResize (static_cast <ResizableInterface *> (tw), x, y, width, height);
    }

    tw->onResizeNotification (x, y, width, height);
}

void
TilesScreen::handleStateLoss (TileAssignableInterface *ta,
			      unsigned int             lostState)
{
    if (priv->mTileLayers.size ())
    {
	TileLayerTrackingGrabsInterface::Ptr tp = priv->mTileLayers.front ();

	tp->handleStateLoss (ta, lostState);
    }
}

void
TilesScreen::dragGrabbedWindow (ResizableInterface *tw,
				TileAssignableInterface *ta,
				WindowMonitorInterface *wm,
				DrawableWindowInterface *dwi,
				MaximizationStateApplicationInterface *max,
				const boost::shared_ptr <AnimatableRectInterface> &animatable,
				const boost::shared_ptr <DragAndDropInterface> &dnd,
				const boost::shared_ptr <ResizeAllocationInterface> &resizeAllocator,
				const CompPoint &pointer)
{
}

namespace
{

bool
dropCommitInterfaceInDraggedWindowInfo (const DraggedWindowInfo &info, DropCommitInterface *tw)
{
    return info.dropCommit == tw;
}

}

void
TilesScreen::removeGrabbedWindow (DropCommitInterface *tw)
{
    auto it = std::find_if (priv->mDraggedWindows.begin (),
			    priv->mDraggedWindows.end (),
			    boost::bind (dropCommitInterfaceInDraggedWindowInfo, _1, tw));

    if (it != priv->mDraggedWindows.end ())
    {
	priv->mDraggedWindows.erase (it);

	/* No more dragged windows and the last tile layer is empty,
	 * delete it */
	tw->dropCommit ();

	if (priv->mTileLayers.size () == 1)
	    if (priv->mTileLayers.front ()->tiles ().empty ())
		priv->mTileLayers.clear ();
    }
    else
    {
	auto rit = std::find (priv->mResizedWindows.begin (), priv->mResizedWindows.end (), tw);

	if (rit != priv->mResizedWindows.end ())
	    priv->mResizedWindows.erase (rit);

	if (priv->mResizedWindows.empty () &&
	    !priv->mTileLayers.empty ())
	{
	    for (const TileInterface::Ptr &tp : priv->mTileLayers.front ()->tiles ())
		tp->lockTargetSize ();

	    priv->cScreen->preparePaintSetEnabled (this, true);
	    priv->cScreen->donePaintSetEnabled (this, true);
	}
    }
}

void
TilesScreen::handleEvent (XEvent *event)
{
    CompWindow *w = NULL;

    switch (event->type)
    {
    case ClientMessage:

	w = screen->findWindow (event->xproperty.window);

	if (w)
	{
	    if (event->xclient.message_type == XInternAtom (screen->dpy (), "_COMPIZ_RESIZE_NOTIFY", 0))
	    {
		TilesScreen::get (screen)->handleWindowResized (TilesWindow::get (w),
								static_cast <int> (event->xclient.data.l[0]),
								static_cast <int> (event->xclient.data.l[1]),
								static_cast <int> (event->xclient.data.l[2]),
								static_cast <int> (event->xclient.data.l[3]));
	    }
	}
    }

    if (event->type == priv->cScreen->damageEvent () + XDamageNotify)
    {
	XDamageNotifyEvent *de = (XDamageNotifyEvent *) event;
	w = screen->findWindow (de->drawable);

	if (w)
	    TilesWindow::get (w)->queueUpdate ();
    }

    screen->handleEvent (event);
}

GrabTrackingInterface *
TilesScreen::findTracker (WindowMonitorInterface *wm, const CompPoint &pointer)
{
    TileLayerTrackingGrabsInterface::Ptr tp;

    /* If there is no layers then we should create one */
    if (priv->mTileLayers.empty ())
    {
	tp.reset (new UnityTileLayer (this));
	/*wm->monitorRegion ().boundingRect (),
				 boost::bind (&WindowMonitorInterface::monitorWorkArea, wm),
				 this));*/

	priv->mTileLayers.push_back (tp);
    }
    else
	tp = priv->mTileLayers.back ();

    return tp.get ();
}

const CompPoint &
TilesScreen::getPointerPosition ()
{
    priv->mPointerPosition = CompPoint (pointerX, pointerY);
    return priv->mPointerPosition;
}

bool
TilesScreen::grabbedWindows (unsigned int mask)
{
    bool ret = false;

    if (mask & CompWindowGrabMoveMask)
	ret |= !priv->mDraggedWindows.empty ();

    if (mask & CompWindowGrabResizeMask)
	ret |= !priv->mResizedWindows.empty ();

    return ret;
}

void
TilesScreen::addResizingWindow (DropCommitInterface *tw)
{
    priv->mResizedWindows.push_back (tw);

    /* this is more or less of a hack to prevent
     * any animations by disabling the preparePaint
     * function and doing the size shifting outselves,
     * can't really think of a cleaner way to do this */
    priv->cScreen->preparePaintSetEnabled (this, false);
    priv->cScreen->donePaintSetEnabled (this, false);
}

bool
TilesScreen::addGrabbedWindow (TileAssignableInterface *assignable,
			       AnimatableWindowInterface *animatable,
			       WindowMonitorInterface   *monitor,
			       DrawableWindowInterface *drawable,
			       DropCommitInterface     *dropCommit,
			       const TileLayer::GetWorkAreaFunc &func,
			       const CompPoint                  &dragPoint)
{
    DraggedWindowInfo info =
    {
	assignable,
	animatable,
	monitor,
	drawable,
	dropCommit
    };

    if (std::find_if (priv->mDraggedWindows.begin (),
		      priv->mDraggedWindows.end (), boost::bind (dropCommitInterfaceInDraggedWindowInfo, _1, dropCommit)) == priv->mDraggedWindows.end ())
    {
	priv->mDraggedWindows.push_back (info);
	return true;
    }

    return false;
}

void
TilesScreen::preparePaint (int ms)
{
    bool continueAnimation = false;

    for (TileLayerTrackingGrabsInterface::Ptr p : priv->mTileLayers)
	if (p->animate (0))
	    continueAnimation |= p->animate (ms);

    for (const DraggedWindowInfo &dwi : priv->mDraggedWindows)
	if (dwi.assignable->tile () && !dwi.assignable->tile ()->layer ())
	    if (dwi.animatable->animate (0))
		continueAnimation |= dwi.animatable->animate (ms);

    /* Update fbos */
    priv->mFramebufferUpdateQueue.dispatch ();
    priv->cScreen->donePaintSetEnabled (this, true);

    priv->cScreen->preparePaint (ms);
}

bool
TilesScreen::glPaintOutput (const GLScreenPaintAttrib &attrib, const GLMatrix &matrix,
			    const CompRegion &region, CompOutput *output, unsigned int mask)
{
    bool transformedWindows = false;

    /* continueAnimation means that there are scaled windows
     * around so we need to allow painting with custom vertex
     * layouts */
    for (const TileLayerTrackingGrabsInterface::Ptr &p : priv->mTileLayers)
    {
	for (const TileInterface::Ptr &tp : p->tiles ())
	{
	    if (!veceq (tp->scale (tp->getDrawWindowGeometry ()), GLVector (1.0, 1.0, 1.0, 1.0)) ||
		!veceq (tp->translation (tp->getDrawWindowGeometry ()), GLVector (0.0, 0.0, 0.0, 0.0)))
	    {
		transformedWindows = true;
		break;
	    }
	}

	if (transformedWindows)
	    break;
    }

    if (transformedWindows)
	mask |= PAINT_SCREEN_WITH_TRANSFORMED_WINDOWS_MASK;

    priv->mCurrentlyPaintedOutput = output;

    return priv->gScreen->glPaintOutput (attrib, matrix, region, output, mask);
}

void
TilesScreen::donePaint ()
{
    bool continueAnimation = false;

    for (const TileLayerTrackingGrabsInterface::Ptr &p : priv->mTileLayers)
    {
	if (p->animate (0))
	{
	    p->damage ();
	    continueAnimation = true;
	}
    }

    for (const DraggedWindowInfo &info : priv->mDraggedWindows)
    {
	if (info.animatable->animate (0))
	    continueAnimation = true;
	else
	{
	    /* Tile not associated with any layers,
	     * delete the tile */
	    if (info.assignable->tile () && !info.assignable->tile ()->layer ())
		info.assignable->releaseTile ();
	}
    }

    priv->cScreen->donePaint ();

    continueAnimation |= priv->mFramebufferUpdateQueue.pending ();

    if (!continueAnimation)
    {
	priv->cScreen->preparePaintSetEnabled (this, false);
	priv->cScreen->donePaintSetEnabled (this, false);
    }
}

PrivateTilesWindow::PrivateTilesWindow (FramebufferUpdateQueueInterface *uq,
					FramebufferUpdateInterface      *ui,
					WindowMonitorInterface          *wm,
					ResizableInterface              *rs,
					MaximizableInterface            *mi) :
    mFramebufferForwarder (boost::bind (&FramebufferUpdateInterface::fboUpdate, ui)),
    mFramebufferCommunicator (uq, &mFramebufferForwarder),
    mMaximizer (rs, wm, mi)
{
}


TilesWindow::TilesWindow (CompWindow *w) :
    PluginClassHandler <TilesWindow, CompWindow> (w),
    priv (new PrivateTilesWindow (TilesScreen::get (screen),
				  this,
				  this,
				  this,
				  this))
{
    priv->window = w;
    priv->cWindow = CompositeWindow::get (w);
    priv->gWindow = GLWindow::get (w);

    priv->mSavedSizeHintsFlags = 0;

    WindowInterface::setHandler (priv->window);
    CompositeWindowInterface::setHandler (priv->cWindow);
    GLWindowInterface::setHandler (priv->gWindow, false);

    priv->window->moveNotifySetEnabled (this, false);

    setFramebufferHolder (TilesScreen::get (screen));

    priv->mGrabTrackerFinder = TilesScreen::get (screen);
    priv->mPointerFetch = TilesScreen::get (screen);
    priv->mGlPaintEnabled = false;
}

TilesWindow::~TilesWindow ()
{
}

TilesScreen::~TilesScreen ()
{
    delete priv->mHPassProgram;
    delete priv->mVPassProgram;
    delete priv;
}

TilesScreen::TilesScreen (CompScreen *s) :
    PluginClassHandler <TilesScreen, CompScreen> (s),
    priv (new PrivateTilesScreen)
{
    priv->cScreen = CompositeScreen::get (screen);
    priv->gScreen = GLScreen::get (screen);

    ScreenInterface::setHandler (screen);
    CompositeScreenInterface::setHandler (priv->cScreen);
    GLScreenInterface::setHandler (priv->gScreen);

    GL::glUniform1fvP = reinterpret_cast <GL::GLUniform1fvProc> (priv->gScreen->getProcAddress ("glUniform1fv"));

    CompString bvp (PrivateTilesScreen::blurVertexProgram);
    CompString hfp (PrivateTilesScreen::hPassFragmentProgram);
    CompString vfp (PrivateTilesScreen::vPassFragmentProgram);

    priv->mHPassProgram = new GLProgram (bvp, hfp);
    priv->mVPassProgram = new GLProgram (bvp, vfp);

    bool usable = priv->mHPassProgram->valid () &&
		  priv->mVPassProgram->valid ();

    compLogMessage ("tiles", CompLogLevelInfo, "fragment program usage: %s", usable ? "yes" : "no");

    priv->mCurrentlyPaintedOutput = NULL;
}


bool
TilesPluginVTable::init ()
{
    if (!CompPlugin::checkPluginABI ("core", CORE_ABIVERSION) ||
	!CompPlugin::checkPluginABI ("composite", COMPIZ_COMPOSITE_ABI) ||
	!CompPlugin::checkPluginABI ("opengl", COMPIZ_OPENGL_ABI))
    {
	return false;
    }

    return true;
}


