/*
 * Compiz Tiles Plugin
 *
 * tiles.h
 *
 * Copyright (c) 2011 Sam Spilsbury <smspillaz@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include <core/core.h>
#include <composite/composite.h>
#include <opengl/opengl.h>
#include <boost/shared_ptr.hpp>

#include "tiles_options.h"

#include "fbeffect.h"
#include "fbmanager.h"
#include "animatable.h"
#include "animatedrects.h"
#include "monitor.h"
#include "geometrysaverestore.h"
#include "resizable.h"
#include "drawable.h"
#include "dnd.h"
#include "resizeallocator.h"
#include "pointerfetch.h"
#include "grabtracker.h"
#include "tile.h"
#include "tilelayer.h"
#include "tilelayerlock.h"
#include "maximization.h"

class TileLayer;
class PrivateTilesScreen;
class PrivateTilesWindow;
typedef boost::shared_ptr <TileLayer> TileLayerPtr;

class TileAssignableInterface;
class TileLayerMiscInterface;
class PointerFetchInterface;

class AnimatedPaintableInterface
{
    public:

    	virtual ~AnimatedPaintableInterface () {}

    	virtual CompRect geometryWithBorders () = 0;
    	virtual CompRect geometryWithShadows () = 0;

    	virtual CompRect tweenedGeometry (const CompRect &rect) = 0;
};

class TilesWindow :
    public PluginClassHandler <TilesWindow, CompWindow>,
    public WindowInterface,
    public CompositeWindowInterface,
    public GLWindowInterface,
    public FramebufferUpdateInterface,
    public FramebufferQueuedUpdateInterface,
    public ResizableInterface,
    public AnimatableWindowInterface,
    public DrawableWindowInterface,
    public TileAssignableInterface,
    public WindowMonitorInterface,
    public DropCommitInterface,
    public GrabTrackableInterface,
    public MaximizableInterface,
    public MaximizationStateApplicationInterface
{
    public:

	TilesWindow (CompWindow *w);
	~TilesWindow ();

	void grabNotify (int x, int y, unsigned int state, unsigned int mask);
	void ungrabNotify ();

	void stateChangeNotify (unsigned int lastState);

	void moveNotify (int dx, int dy, bool immediate);
	void validateResizeRequest (unsigned int &mask,
				    XWindowChanges *xwc,
				    unsigned int source);
	void onResizeNotification (int x, int y, int width, int height);

	bool glPaint (const GLWindowPaintAttrib &attrib, const GLMatrix &matrix,
		      const CompRegion &region, unsigned int mask);
	bool damageRect (bool initial, const CompRect &rect);
	void windowNotify (CompWindowNotify n);

	bool animate (unsigned int ms);

	void damage ();

	bool assignTile (TileInterface::Ptr tp);
	void releaseTile ();

	void dropCommit ();

	void setGrabTracker (GrabTrackingInterface *grabTracker);

	const TileInterface::Ptr & tile () const;

	CompRect grabGeometry ();
	CompRect getDrawWindowGeometry ();

	void queueUpdate ();
	void dequeueUpdates ();

	CompRect frameGeometry ();
	CompRect windowGeometry ();
	void resizeWindow (const CompRect &r);

	CompRect toWindowGeometry (const CompRect &r);
	CompRect toFrameGeometry (const CompRect &r);

	void applyMaximizeState (const CompRect &resizeTarget);
	void applyResizeRules (bool lock);

	void glDrawSetEnabled (bool);

	CompRegion monitorRegion ();
	CompRect monitorWorkArea ();

	void maximize (unsigned int);

    private:

    	bool fboUpdate ();
    	void setFramebufferHolder (FramebufferUpdateQueueInterface *updateQueue);

	PrivateTilesWindow *priv;
};

class TilesScreen :
    public PluginClassHandler <TilesScreen, CompScreen>,
    public ScreenInterface,
    public CompositeScreenInterface,
    public GLScreenInterface,
    public TilesOptions,
    public FramebufferUpdateQueueInterface,
    public GrabTrackingInterface,
    public GrabTrackerFinderInterface,
    public PointerFetchInterface
{
    public:

	TilesScreen (CompScreen *);
	~TilesScreen ();

	void preparePaint (int);
	bool glPaintOutput (const GLScreenPaintAttrib &attrib, const GLMatrix &matrix,
			    const CompRegion &region, CompOutput *output, unsigned int mask);
	void donePaint ();

	bool addGrabbedWindow (TileAssignableInterface *assignable,
			       AnimatableWindowInterface *animatable,
			       WindowMonitorInterface   *monitor,
			       DrawableWindowInterface *drawable,
			       DropCommitInterface *dropCommit,
			       const TileLayer::GetWorkAreaFunc &func,
			       const CompPoint &dragPoint);
	void removeGrabbedWindow (DropCommitInterface *tw);
	void addResizingWindow (DropCommitInterface *tw);
	void dragGrabbedWindow (ResizableInterface *tw,
				TileAssignableInterface *ta,
				WindowMonitorInterface *wm,
				DrawableWindowInterface *dwi,
				MaximizationStateApplicationInterface *max,
				const boost::shared_ptr <AnimatableRectInterface> &animatable,
				const boost::shared_ptr <DragAndDropInterface> &dnd,
				const boost::shared_ptr <ResizeAllocationInterface> &resizeAllocator,
				const CompPoint &pointer);
	void handleWindowResized (TilesWindow *tw, int x, int y, int width, int height);
	bool grabbedWindows (unsigned int);

	void handleStateLoss (TileAssignableInterface *,
			      unsigned int);

	void handleEvent (XEvent *event);

	GrabTrackingInterface * findTracker (WindowMonitorInterface *wm, const CompPoint &pointer);
	const CompPoint & getPointerPosition ();

	GLProgram * vPassProgram () const;
	GLProgram * hPassProgram () const;
	CompOutput * currentlyPaintedOutput () const;

    private:

    	bool addQueuedUpdate (FramebufferUpdateInterface *update);
	void removeUpdatesFromQueue (FramebufferUpdateInterface *update);
	void dispatch ();
	bool pending ();

	PrivateTilesScreen *priv;
};

class TilesPluginVTable :
    public CompPlugin::VTableForScreenAndWindow<TilesScreen, TilesWindow>
{
    public:

	bool init ();
};
